"""
Dependency injectors make communicating with the routines in `psf_data_ops.py` more managable
by providing a consistent interface. As long as the dependency injector provides parameters
and functions as specified in this file, the fitting functions should work.

"""








from typing import TypeVar, TypeVarTuple, ParamSpec, Generic, Callable, Any, NewType, Protocol

import numpy as np

from aopp_deconv_tool.optimise_compat import PriorParam, PriorParamSet

from aopp_deconv_tool.gaussian_psf_model import GaussianPSFModel
from aopp_deconv_tool.radial_psf_model import RadialPSFModel
from aopp_deconv_tool.turbulence_psf_model import TurbulencePSFModel, SimpleTelescope, CCDSensor
from aopp_deconv_tool.optics.turbulence_model import phase_psd_von_karman_turbulence as turbulence_model
from aopp_deconv_tool.adaptive_optics_psf_model import PSFModel as AOInstrumentPSFModel
from aopp_deconv_tool.optics.turbulence_model import phase_psd_von_karman_turbulence
from aopp_deconv_tool.optics.adaptive_optics_model import phase_psd_fetick_2019_moffat_function
from aopp_deconv_tool.instrument_model.vlt import VLT


T = TypeVar('T')
Ts = TypeVarTuple('Ts')

# Integers that may or may not be equal
type N = TypeVar('N',bound=int)
type M = TypeVar('M',bound=int)

# A length N tuple of integers
type S[N] = GenericAlias(tuple, (int,)*N)






# Need to have some way to describe how to communicate with the dependency injection mechanism,
# this isn't the best, but it's what I could come up with. 
# `ParamsAndPsfModelDependencyInjector` is a base class that defines the interface
# I could probably do this with protocols, but I think it would be harder to communicate my intent

# We don't know the shape of the PSF data, but it must be a have two integer values
type Ts_PSF_Data_Array_Shape = S[Literal[2]]

# we require that this argument set is compatible with parameters specified in a 'PriorParamSet' instance
type P_ArgumentsLikePriorParamSet = ParamSpec('ArgumentsLikePriorParamSet')

# PSF data is a numpy array of some specified shape and type 'T'
type T_PSF_Data_NumpyArray = np.ndarray[Ts_PSF_Data_Array_Shape, T] 

# We want the callable we are given to accept parameters in a way that is compatible with 'PriorParamSet', and return a numpy array that is like our PSF Data
type T_PSF_Model_Flattened_Callable = Callable[P_ArgumentsLikePriorParamSet, T_PSF_Data_NumpyArray] 

# Fitted varaibles from `psf_data_ops.fit_to_data(...)` are returned as a dictionary
type T_Fitted_Variable_Parameters = dict[str,Any] 

# Constant paramters to `psf_data_ops.fit_to_data(...)` are returned as a dictionary
type T_Constant_Parameters = dict[str,Any] 

# If we want to postprocess the fitted PSF result, we will need to know the PriorParamSet used, the callable used, the fitted variables, and the constant paramters resulting from the fit.
type P_PSF_Result_Postprocessor_Arguments = [PriorParamSet, T_PSF_Model_Flattened_Callable, T_Fitted_Variable_Parameters, T_Constant_Parameters] 

# If we preprocess the fitted PSF, we must return something that is compatible with the PSF data.
type T_PSF_Result_Postprocessor_Callable = Callable[
	P_PSF_Result_Postprocessor_Arguments, 
	T_PSF_Data_NumpyArray
]

# A callable that accepts a dictionary of fitted parameter values
type T_Fitted_Parameter_Callable = Callable[[dict[str,float], ...], T_PSF_Data_NumpyArray] 




def prior_param_args_from_param_spec(
		param_name : str, 
		is_const_default : bool | None, 
		initial_value_default : float | None, 
		range_default : tuple[float,float] | None, 
		var_params : list[str] | tuple[str], 
		const_params : list[str] | tuple[str], 
		initial_values : dict[str, float], 
		range_values : dict[str, tuple[float,float]]
	) -> tuple[str, tuple[float,float], bool, float]:
	
	if (is_const_default == None) and ((param_name not in var_params) | (param_name not in const_params)):
		raise RuntimeError(f'Parameter "{param_name}" cannot be assumed to be constant or variable, it must be defined as such.')
	if (initial_value_default == None) and (param_name not in initial_values):
		raise RuntimeError(f'Parameter "{param_name}" does not have an initial value assigned and cannot use a default value')
	if (range_default == None) and (param_name not in range_values):
		raise RuntimeError(f'Parameter "{param_name}" does not have a range assigned and cannot use a default range')

	return (
		param_name,
		range_values.get(param_name, range_default),
		False if param_name in var_params else True if param_name in const_params else is_const_default,
		initial_values.get(param_name, initial_value_default)
	)



class ParamsAndPsfModelDependencyInjector:
	"""
	Subclass this and overwrite it's methods to get something that works like something in the ".../examples/psf_model_example.py" script.
	"""
	def __init__(self, psf_data : T_PSF_Data_NumpyArray):
		self.psf_data = psf_data
		self._psf_model = NotImplemented # this will be defined here in a subclass
		self._params = NotImplemented # PriorParamSet(), this will be defined here in a subclass
	
	def get_psf_model_name(self):
		"""
		Returns the name of the PSF model. Defaults to the class name.
		"""
		return self._psf_model.__class__.__name__

	def get_parameters(self) -> PriorParamSet:
		"""
		Returns the PriorParamSet that represents the parameters for the `self._psf_model`
		"""
		return self._params
	
	def get_psf_model_flattened_callable(self) -> T_PSF_Model_Flattened_Callable :
		"""
		Returns a wrapper around `self._psf_model` that accepts all of it's parameters as simple-data arguments (no lists or objects)
		"""
		NotImplemented
	
	def get_psf_result_postprocessor(self) -> None | T_PSF_Result_Postprocessor_Callable :
		"""
		Returns a callable that postprocesses the result from `self._psf_model`
		"""
		NotImplemented
	
	def get_fitted_parameters_callable(self) -> T_Fitted_Parameter_Callable:
		"""
		Return a callable that accepts a dictionary of fitted parameteter values
		"""
		def fitted_parameters_callable(fitted_params : dict[str,float]):
			return self._params.apply_to_callable(self.get_psf_model_flattened_callable(), fitted_params, self._params.consts)
		
		return fitted_parameters_callable


class RadialPSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF as a radial histogram with `nbins` from point (`x`,`y`)
	"""
	
	def __init__(self, psf_data):
		
		super().__init__(psf_data)
		
		self._params = PriorParamSet(
			PriorParam(
				'x',
				(0, psf_data.shape[0]),
				False,
				psf_data.shape[0]//2
			),
			PriorParam(
				'y',
				(0, psf_data.shape[1]),
				False,
				psf_data.shape[1]//2
			),
			PriorParam(
				'nbins',
				(0, np.inf),
				True,
				50
			)
		)
		
		self._psf_model = RadialPSFModel(
			psf_data
		)
		
	
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self): 
		return self._psf_model
	
	def get_psf_result_postprocessor(self): 
		def psf_result_postprocessor(params, psf_model_flattened_callable, fitted_vars, consts):
			params.apply_to_callable(
				psf_model_flattened_callable, 
				fitted_vars,
				consts
			)
			return psf_model_flattened_callable.centered_result
			
		return psf_result_postprocessor


class GaussianPSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF as a 2d gaussian with mean (`x`,`y`), standard deviation `sigma` an offset `const` from zero, and a multiplication factor `factor`
	"""
	
	def __init__(self, psf_data):
		
		super().__init__(psf_data)
		
		self._params = PriorParamSet(
			PriorParam(
				'x',
				(0, psf_data.shape[0]),
				True,
				psf_data.shape[0]//2
			),
			PriorParam(
				'y',
				(0, psf_data.shape[1]),
				True,
				psf_data.shape[1]//2
			),
			PriorParam(
				'sigma',
				(0, np.sum([x**2 for x in psf_data.shape])),
				False,
				5
			),
			PriorParam(
				'const',
				(0, 1),
				False,
				0
			),
			PriorParam(
				'factor',
				(0, 2),
				False,
				1
			)
		)
		
		self._psf_model = GaussianPSFModel(psf_data.shape, float)
	
	
	
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self): 
		def psf_model_flattened_callable(x, y, sigma, const, factor):
			return self._psf_model(np.array([x,y]), np.array([sigma,sigma]), const)*factor
		return psf_model_flattened_callable
	
	def get_psf_result_postprocessor(self): 
		return None


class TurbulencePSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF as the result of von-karman turbulence. Assumes the PSF is at the center of the model.
	"""
	
	def __init__(self, psf_data):
		super().__init__(psf_data)
		
		self._params = PriorParamSet(
			PriorParam(
				'wavelength',
				(0, np.inf),
				True,
				750E-9
			),
			PriorParam(
				'r0',
				(0, 1),
				True,
				0.1
			),
			PriorParam(
				'turbulence_ndim',
				(0, 3),
				False,
				1.5
			),
			PriorParam(
				'L0',
				(0, 50),
				False,
				8
			)
		)
		
		self._psf_model = TurbulencePSFModel(
			SimpleTelescope(
				8, 
				200, 
				CCDSensor.from_shape_and_pixel_size(psf_data.shape, 2.5E-6)
			),
			turbulence_model
		)
		
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self): 
		return self._psf_model
	
	def get_psf_result_postprocessor(self): 
		return None



class MUSEAdaptiveOpticsPSFModelDependencyInjector(ParamsAndPsfModelDependencyInjector):
	"""
	Models the PSF using a model of the adaptive optics MUSE instrument on the VLT telescope.
	"""
	
	def __init__(self, 
			psf_data, 
			var_params : list[str] | tuple[str] = [], 
			const_params : list [str] | tuple [str] = [],
			initial_values : dict[str,float] = {},
			range_values : dict[str,tuple[float,float]] = {}
		):
		super().__init__(psf_data)
		
		instrument = VLT.muse(
			expansion_factor = 3,
			supersample_factor = 2,
			obs_shape=psf_data.shape[-2:]
		)
		
		self._psf_model = AOInstrumentPSFModel(
			instrument.optical_transfer_function(),
			phase_psd_von_karman_turbulence,
			phase_psd_fetick_2019_moffat_function,
			instrument
		)
		
		self._params = params = PriorParamSet(
			PriorParam(
				*prior_param_args_from_param_spec('wavelength', True, 750E-9, (0,np.inf), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('r0', True, 0.15, (0.01,10), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('turb_ndim', True, 1.3, (1,2), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('L0', True, 1.5, (0,10), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('alpha', False, 0.4, (0.1,3), var_params, const_params, initial_values, range_values) 
			),
			PriorParam(
				*prior_param_args_from_param_spec('beta', True, 1.6, (1.1, 10), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('ao_correction_frac_offset', False, 0, (-1,1), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('ao_correction_amplitude', False, 2.2, (0,5), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('factor', False, 1, (0.7,1.3), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('s_factor', True, 0, (0,np.inf), var_params, const_params, initial_values, range_values)
			),
			PriorParam(
				*prior_param_args_from_param_spec('f_ao', True, instrument.f_ao, (24.0/(2*instrument.obj_diameter),52.0/(2*instrument.obj_diameter)), var_params, const_params, initial_values, range_values)
			)
		)
		
		
		
	def get_parameters(self):
		return self._params
	
	def get_psf_model_flattened_callable(self):
		parent = self
		class PSFModelFlattenedCallable:
			def __init__(self):
				self.specific_model = None
				self.result = None
				self.cached_args = np.zeros((10,))
				
			
			def __call__(
					self,
					wavelength,
					r0, 
					turb_ndim, 
					L0, 
					alpha, 
					beta,
					f_ao,
					ao_correction_amplitude, 
					ao_correction_frac_offset, 
					s_factor,
					factor
				):

				# if we should recalculate everything, do so. Otherwise use the saved model
				if np.any(np.abs(np.array((r0, turb_ndim, L0, alpha, beta, f_ao,ao_correction_amplitude, ao_correction_frac_offset, s_factor,factor)) - self.cached_args) > 1E-5):
					self.specific_model = parent._psf_model(
						None,
						(r0, turb_ndim, L0),
						(alpha, beta),
						f_ao,
						ao_correction_amplitude,
						ao_correction_frac_offset,
						s_factor
					)
				
				
				#_lgr.debug(f'{(wavelength,r0, turb_ndim, L0, alpha, beta, f_ao,ao_correction_amplitude, ao_correction_frac_offset, s_factor,factor)}')
				self.result = self.specific_model.at(wavelength, plots=False).data
				
				return factor*(self.result / np.nansum(self.result.data))
		
		return PSFModelFlattenedCallable()
	
	def get_psf_result_postprocessor(self): 
		def psf_result_postprocessor(params, psf_model_flattened_callable, fitted_vars, consts):
			result = params.apply_to_callable(
				psf_model_flattened_callable, 
				fitted_vars,
				consts
			)
			return result
			
		return psf_result_postprocessor




"""
Routines that use masks to assist in array operations
"""
import numpy as np

from aopp_deconv_tool.geometry.shape import GeoShape
from aopp_deconv_tool.numpy_helper.array import N,S

def from_nan_and_inf(a : np.ndarray[[...],float]) -> np.ndarray[[...],bool]:
	"""
	Gets a mask of all nan and inf values

	Arguments:
		a : np.ndarray
			A numpy array of pixels
	
	Returns:
		A boolean numpy array that is True where `a` is NAN and INF.
	"""
	return np.isnan(a) | np.isinf(a)

def from_shape(shape : S[N], geo_shape : GeoShape, extent : np.ndarray[[N,2],float] | None = None) -> np.ndarray[S[N],bool]:
	"""
	Gets a mask of `shape` and start-end coords `extent` that are inside `poly_shape`
	"""
	if extent is None:
		# Assume we want to center the geo_shape in the region
		extent = np.array(tuple((-(s-1)/2, (s-1)/2) for s in shape))
	print(f'{extent=}')
	print(f'{extent[:,1]=}')
	stride = (extent[:,1]-extent[:,0]) / np.array(tuple(s-1 for s in shape)) # does not include end point
	print(f'{stride=}')
	#return np.vectorize(lambda p: p in geo_shape)(np.moveaxis(np.indices(shape),0,-1)*stride + extent[:,0])
	print(f'{np.moveaxis(np.moveaxis(np.indices(shape),0,-1)*stride + extent[:,0], -1,0)}')
	return geo_shape.__contains__(np.moveaxis(np.indices(shape),0,-1)*stride + extent[:,0])





import numpy as np


class RadialPSFModel:
	"""
	PSF model that is a radially averaged PSF (i.e. a histogram of radial bins).
	"""
	def __init__(self, data):
		self.data = data
		self.r_bins = np.zeros((1,2))
		self.r_values = np.zeros((1,))
		self._result = np.zeros_like(data)
	
	def __call__(self, x, y, nbins):
		r_max = max(*self.data.shape)
		nbins = int(nbins)
		self.r_values = np.zeros((nbins,))
		self.r_bins = np.zeros((nbins,2))
		self.r_bins[:,1] = np.logspace(0, np.log10(r_max), nbins)
		self.r_bins[1:,0] = self.r_bins[:-1,1]
		
		r = np.sqrt(np.sum((np.indices(self.data.shape).T - np.array([x,y]))**2, axis=-1)).T
		
		self._result *= 0
		for i, r_lim in enumerate(self.r_bins):
			r_mask = (r_lim[0] <= r) & (r < r_lim[1])
			self.r_values[i] = np.nanmean(self.data[r_mask])
			self._result[r_mask] = self.r_values[i]
		
		return self._result
	
	
	@property
	def centered_result(self):
		r = np.sqrt(np.sum((np.indices(self.data.shape).T - np.array([self.data.shape[0]//2, self.data.shape[1]//2]))**2, axis=-1)).T
		
		self._result *= 0
		for i, r_lim in enumerate(self.r_bins):
			r_mask = (r_lim[0] <= r) & (r < r_lim[1])
			self._result[r_mask] = self.r_values[i]
		return self._result


"""
Quick tool for interpolating data in a FITS file
"""


import sys
from pathlib import Path

from typing import Literal

import numpy as np
import scipy as sp
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice

import aopp_deconv_tool.numpy_helper.array.grid

import aopp_deconv_tool.scipy_helper as sph
import aopp_deconv_tool.scipy_helper.interp

from aopp_deconv_tool.algorithm.interpolate.ssa_interp import ssa_intepolate_at_mask
from aopp_deconv_tool.algorithm.bad_pixels.ssa_sum_prob import ssa2d_sum_prob_map

from aopp_deconv_tool.py_ssa import SSA

import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')



data_ssa = None

def get_data_ssa(a, **kwargs):
	global data_ssa
	if data_ssa is None:
		data_ssa = SSA(
			a, 
			**{
				'w_shape':20,
				#'grouping' : {'mode' : 'similar_eigenvalues', 'tolerance' : 1E-1},
				**kwargs
			},
		)
	return data_ssa

def set_data_ssa(value):
	global data_ssa
	data_ssa = value

def get_bad_pixel_map(
		a : np.ndarray, 
		method : Literal['ssa'] | Literal['simple'], 
		perform_binary_operation : None | list[Literal['opening'] | Literal['closing'] | Literal['dilation'] | Literal['erosion']] = None,
	) -> np.ndarray:
	nan_inf_mask = np.isinf(a) | np.isnan(a)
	match method:
		case 'ssa':
			ssa = get_data_ssa(np.nan_to_num(a))
			bad_pixel_map = ssa2d_sum_prob_map(
				ssa, 
				value=1, 
				start=5, 
				stop=None, 
				strategy='n_std_dev_from_median',
				transform_value_as='identity',
				show_plots=0
			)
		case 'simple':
			bad_pixel_map = nan_inf_mask
		case _:
			raise RuntimeError(f'Unknown value of {method=}')
	
	
	if perform_binary_operation is not None:
		orig_bp_map = np.array(bad_pixel_map, dtype=bad_pixel_map.dtype)
		for item in perform_binary_operation:
			match item:
				case 'closing':
					bad_pixel_map = sp.ndimage.binary_closing(bad_pixel_map)
				case 'opening':
					bad_pixel_map = sp.ndimage.binary_opening(bad_pixel_map)
				case 'erosion':
					bad_pixel_map = sp.ndimage.binary_erosion(bad_pixel_map)
				case 'dilation':
					bad_pixel_map = sp.ndimage.binary_dilation(bad_pixel_map)
				case 'remove_single_pixels':
					labels, n_labels = sp.ndimage.label(bad_pixel_map)
					reject_labels = np.zeros_like(bad_pixel_map, dtype=bool)
					for i in range(n_labels):
						lbl_mask = labels==i
						n_pixels = np.count_nonzero(lbl_mask)
						if n_pixels <= 1:
							reject_labels[lbl_mask] = True
					bad_pixel_map[reject_labels] = False
					
				case _:
					raise RuntimeError(f'Unknown binary operation "{item}"')
	
	
	return bad_pixel_map | nan_inf_mask

def get_interp_at_mask(a, mask, method):
	match method:
		case 'ssa':
			interp = ssa_intepolate_at_mask(get_data_ssa(a), mask)
		case 'scipy':
			interp = sph.interp.interpolate_at_mask(a, mask, edges='convolution')
		case _:
			raise RuntimeError(f'Unknown value of {method=}')
	
	return interp
	


def run(
		fits_spec,
		output_path,
		bad_pixel_method = 'ssa',
		interp_method = 'ssa',
	):
	
	
	
	
	with fits.open(Path(fits_spec.path)) as data_hdul:
		
		
		_lgr.debug(f'{fits_spec.path=} {fits_spec.ext=} {fits_spec.slices=} {fits_spec.axes=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data
		
		bad_pixel_map = np.zeros_like(data, dtype=bool)
		interp_data = np.full_like(data, fill_value=np.nan)
		
		bad_pixel_map_binary_operations = ['closing', 'remove_single_pixels']
		
		# Loop over the index range specified by `obs_fits_spec` and `psf_fits_spec`
		for i, idx in enumerate(nph.slice.iter_indices(data, fits_spec.slices, fits_spec.axes['CELESTIAL'])):
			_lgr.debug(f'{i=}')
			set_data_ssa(None)
			
			#plt.imshow(data[idx])
			#plt.show()
			
			bad_pixel_map[idx] = get_bad_pixel_map(data[idx], bad_pixel_method, bad_pixel_map_binary_operations)
			#plt.imshow(bad_pixel_map[idx])
			#plt.show()
			
			interp_data[idx] = get_interp_at_mask(data[idx], bad_pixel_map[idx], interp_method)
			#plt.imshow(interp_data[idx])
			#plt.show()
	
	
		hdr = data_hdu.header
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'bad_pixel_method' : bad_pixel_method,
			'interp_method' : interp_method,
		}
		for i, x in enumerate(bad_pixel_map_binary_operations):
			param_dict[f'bad_pixel_map_binary_operations_{i}'] = x
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='interpolate',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
				

	
	# Save the products to a FITS file
	hdu_interp = fits.PrimaryHDU(
		header = hdr,
		data = interp_data
	)
	hdu_bad_pixels = fits.ImageHDU(
		data = bad_pixel_map.astype(int),
		header = hdr,
		name='BAD_PIXELS'
	)
	hdul_output = fits.HDUList([
		hdu_interp,
		hdu_bad_pixels,
	])
	hdul_output.writeto(output_path, overwrite=True)
	

def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter
	)
	
	parser.add_argument(
		'fits_spec', 
		help = aopp_deconv_tool.text.wrap(
			aph.fits.specifier.get_help(['CELESTIAL']).replace('\t', '    '),
			os.get_terminal_size().columns - 30
		)
	)
	parser.add_argument('-o', '--output_path', help='Output fits file path. By default is same as fie `fits_spec` path with "_interp" appended to the filename')
	
	parser.add_argument('--bad_pixel_method', choices=['ssa', 'simple'], default='ssa', help='Strategy to use when finding bad pixels to interpolate over')
	#parser.add_argument('--bad_pixel_args', nargs='*', help='Arguments to be passed to `--bad_pixel_method`')

	parser.add_argument('--interp_method', choices=['ssa', 'scipy'], default='scipy', help='Strategy to use when interpolating')


	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, ['CELESTIAL'])
	
	if args.output_path is None:
		args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+'_interp'+str(Path(args.fits_spec.path).suffix)))
	
	return args


if __name__ == '__main__':
	args = parse_args(sys.argv[1:])
	
	run(args.fits_spec, bad_pixel_method=args.bad_pixel_method, interp_method=args.interp_method, output_path=args.output_path)
	



"""
Helper functions for axes operations on numpy arrays.

Axes referrs to the indexed dimensions of a numpy array. I.e. `a.ndim` is the number
of dimensions in a, and also the number of axes. `a.shape` is the length of each
axis.
"""
import typing
import numpy as np
from contextlib import contextmanager

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARN')



class AxesOrdering:
	"""
	A class that can shift between numpy, fits, and fotran axes ordering
	"""
	__slots__ = (
		'_idx',
		'_n'
	)
	def __init__(self, 
			idx : int, 
 			n : int, 
			ordering : typing.Literal['numpy','fits','fortran']
		) -> None:
		self.n = n
		setattr(self, ordering, idx)
		return

	@classmethod
	def range(cls, start, stop=None, step=1):
		# make interface work like "range()" command
		if stop is None:
			stop = start
			start = 0
		return(cls(x, stop-start, 'numpy') for x in range(start,stop,step))

	@property
	def n(self) -> int:
		return(self._n)
	@n.setter
	def n(self, value: int) -> None:
		self._n = value
		return

	# numpy representation is the internal one, so just return attribute
	@property
	def numpy(self) -> int:
		return(self._idx)
	@numpy.setter
	def numpy(self,value:int)->None:
		self._idx = value
		return

	@property
	def fits(self)->int:
		if isinstance(self._idx, typing.Sequence):
			return(type(self._idx)([self._n - _i for _i in self._idx]))
		return(self._n - self._idx)
		
	@fits.setter
	def fits(self,value:int)->None:
		# convert to numpy representation and store
		if isinstance(value, typing.Sequence):
			self._idx = type(value)([self._n - _v for _v in value])
		else:
			self._idx = self._n - value
		return

	@property
	def fortran(self)->int:
		if isinstance(self._idx, typing.Sequence):
			return(type(self._idx)([1+_i for _i in self._idx]))
		return(1+self._idx)
		
	@fortran.setter
	def fortran(self, value:int)->None:
		if isinstance(value, typing.Sequence):
			self._idx = type(value)([_v-1 for _v in value])
		else:
			self._idx = value-1
		return
# END class AxesOrdering





@contextmanager
def to_end(a : np.ndarray, axes : tuple[int,...]):
	"""
	Reorders the axes of `a` so that the axes specified in the `axes` tuple are 
	at the end (rhs) of the ordering for purposes of slicing and indexing.
	"""
	axes_dest = tuple(a.ndim - 1 -i for i in range(len(axes)))[::-1]
	
	yield np.moveaxis(a, axes, axes_dest)
	
	# as np.moveaxis returns a view of `a` we don't have to clean up anything

@contextmanager
def to_start(a : np.ndarray, axes : tuple[int,...]):
	"""
	Reorders the axes of 'a' so the axes specified in the 'axes' tuple are
	at the start (lhs) of the ordering for slicing and indexing
	"""
	axes_dest = tuple(range(len(axes)))
	
	yield np.moveaxis(a, axes, axes_dest), axes_dest


def iter_axes_group(
		a : np.ndarray,
		axes : tuple[int,...],
	):
	"""
	Iterate over a group of axes
	"""
	with to_end(a, axes) as b:
		n = b.ndim - len(axes)
		shape = (*b.shape[:-len(axes)],1)
		size = np.product(shape)
		strides = tuple(np.cumprod(shape[::-1])[:-1][::-1])
		steps = [0]*n
		#print(f'{n=} {strides=} {steps=}')
		for i in range(size):
			for j, stride in enumerate(strides):
				#print(f'{i=}')
				step = i//stride
				steps[j] = step
				i -= step*stride
				#print(f'{j=} {step=} {steps=} {i=}')
			yield (tuple(steps), b)
				


def reverse(a : np.ndarray):
	"""
	Reverse the axes of `a`
	"""
	axes = tuple(range(a.ndim))
	return np.moveaxis(a, axes, axes[::-1])

def not_in(a : np.ndarray, group : tuple[int,...]) -> tuple[int,...]:
	"""
	Returns the axes of `a` that are not in `group`
	"""
	return tuple(i for i in range(a.ndim) if i not in group)

def merge(a : np.ndarray, axes : tuple[int,...], axis : int = 0):
	"""
	Merge a set of axes into a single axis.
	
	a : np.ndarray
		Numpy array to group axes of
	axes : tuple[int:...]
		axes to group together
	axis : int
		location of the final grouped axis
	"""
	not_merge_shape = (a.shape[i] for i in range(a.ndim) if i not in axes)
	
	return np.moveaxis(np.moveaxis(a, axes, tuple(range(len(axes)))).reshape(-1, *not_merge_shape), 0, axis)






"""
Definitions of generic types so I can type hint more easily
"""

from typing import TypeVar, GenericAlias, NewType

T = TypeVar('T')

type NTuple[T,N] = GenericAlias(tuple, (T,)*N)

type NumVar = int | float

type ShapeVar[N] = NTuple[int,N]




"""
Routines for handing an array that knows it's physical size.
"""


import itertools as it
import numpy as np
from typing import Callable

from matplotlib import pyplot as plt

class GeoArray:
	"""
	A Geometric Array. An array of data, along with coordinate axes that 
	describe the position of the data.
	
	# ATTRIBUTES #
		self.data : np.ndarray
			The data the array holds
		self.axes : np.array
			The 'points in space' where `self.data` is defined. These are what changes when the GeoArray undergoes a geometric transform
	"""
	def __init__(self,
			data : np.ndarray,
			axes : np.ndarray | None,
		):
		self.data = data
		if axes is None:
			self.axes = tuple(np.linspace(-s/2, s/2, s) for s in self.data.shape())
		else:
			self.axes = axes
		
		
	@staticmethod
	def scale_to_axes(scale : tuple[float,...], shape : tuple[int,...], center : float = 0) -> np.ndarray:
		"""
		Applies a `scale` to a `shape` such that the result is centered around `center`
		"""
		return tuple(
			np.linspace(center-scale/2,center+scale/2,s) for scale, s in zip(scale,shape)
		)
	
	def copy(self):
		"""
		Copy the GeoArray
		"""
		return GeoArray(np.array(self.data), np.array(self.axes))
	
	def __array__(self):
		"""
		For compatibility with numpy routines
		"""
		return self.data
	
	def fft(self):
		"""
		Return the fast fourier transform of the GeoArray
		"""
		return GeoArray(
			np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(self.data))),
			tuple(np.fft.fftshift(np.fft.fftfreq(x.size, x[1]-x[0])) for x in self.axes),
		)
	
	def ifft(self):
		"""
		Return the inverse fast fourier transform of the GeoArray
		"""
		return GeoArray(
			np.fft.fftshift(np.fft.ifftn(np.fft.ifftshift(self.data))),
			tuple(np.fft.fftshift(np.fft.fftfreq(x.size, x[1]-x[0])) for x in self.axes),
		)
	
	@property
	def extent(self):
		"""
		Return the "extent" (i.e. min,max in each dimension) of the GeoArray
		"""
		return tuple(it.chain.from_iterable((x[0],x[-1]) for x in self.axes))

	@property
	def scale(self) -> tuple[float,...]:
		"""
		Return the scale (i.e. difference between min and max) of the GeoArray
		"""
		return tuple(x[-1]-x[0] for x in self.axes)

	@property
	def mesh(self) -> np.ndarray:
		"""
		Return a meshgrid that covers the GeoArray
		"""
		return np.array(np.meshgrid(*self.axes[::-1]))




def plot_ga(
		geo_array : GeoArray, 
		data_mutator : Callable[[np.ndarray], np.ndarray] = lambda x:x, 
		title : str = '', 
		data_units : str = '', 
		axes_units : str | tuple[str,...] = '',
		show : bool = True
	):
	"""
	Plot the GeoArray as an image. Only works for 2d arrays at present. Will call
	a mutator on the data before plotting. Returns a numpy array of the axes
	created for the plot
	"""
	if type(axes_units) is not tuple:
		axes_units = tuple(axes_units for _ in range(geo_array.data.ndim))
	plt.clf()
	f = plt.gcf()
	f.suptitle(title)
	
	if geo_array.data.ndim != 2:
		raise NotImplementedError("Geometric array plotting only works for 2d arrays at present.")
		
	a = f.subplots(2,2,squeeze=True).flatten()
	
	center_idx = tuple(s//2 for s in geo_array.data.shape)
	
	m_data = data_mutator(geo_array.data)
	m_axes = geo_array.axes
	
	a[0].set_title('array data')
	a[0].imshow(m_data, extent=geo_array.extent)
	a[0].set_xlabel(axes_units[0])
	a[0].set_ylabel(axes_units[1])
	
	a[1].set_title('x=constant centerline')
	a[1].plot(m_data[:, center_idx[1]], m_axes[0])
	a[1].set_xlabel(data_units)
	a[1].set_ylabel(axes_units[0])
	
	a[2].set_title('y=constant centerline')
	a[2].plot(m_axes[1], m_data[center_idx[0], :])
	a[2].set_xlabel(axes_units[1])
	a[2].set_ylabel(data_units)
	
	a[3].set_title('array data unmutated')
	a[3].imshow(geo_array.data if geo_array.data.dtype!=np.dtype(complex) else np.abs(geo_array.data), extent=geo_array.extent)
	a[3].set_xlabel(axes_units[0])
	a[3].set_ylabel(axes_units[1])
	
	
	if show: plt.show()
	return f, a



"""
Contains classes and functions that aid compatibility with the ultranest package
"""
import sys
import json
from typing import ParamSpec, TypeVar, Callable, Any
from pathlib import Path

import numpy as np
import ultranest

import aopp_deconv_tool.plot_helper as plot_helper
import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')

from aopp_deconv_tool.optimise_compat import PriorParamSet



T = TypeVar('T')
P = ParamSpec('P')

def fitting_function_factory(
		reactive_nested_sampler_kwargs : dict[str,Any] = {}, 
		sampler_run_kwargs : dict[str,Any] = {}
	) -> Callable[[PriorParamSet, Callable[...,float], list[str]|tuple[str], list[str]|tuple[str]], [...]]:
	"""
	Takes in two dictionaries of keyword arguments for ultranest's "ReactiveNestedSampler" class and "Sampler.run" method,
	returns a callable that accepts a PriorParamSet, a scipy-compatible objective function, a list of variable parameters,
	and a list of constant parameters. This returned function runs the created "ReactiveNestedSampler"'s ".run" method.
	"""
	
	reactive_nested_sampler_kwargs_defaults = dict(
		resume = 'subfolder',
		run_num = 0
	)
	reactive_nested_sampler_kwargs_defaults.update(reactive_nested_sampler_kwargs)
	
	reactive_nested_sampler_kwargs = reactive_nested_sampler_kwargs_defaults
	
	sampler_run_kwargs_defaults = dict(
		max_iters=2000, #500, # 2000,
		max_ncalls=10000, #5000
		frac_remain=1E-2,
		Lepsilon = 1E-1,
		min_num_live_points=200, #20, #80
		cluster_num_live_points=40, #1, #40
		dlogz=100,
		min_ess=200, #1, #40
		update_interval_volume_fraction=0.99, #0.8
		max_num_improvement_loops=3,
		widen_before_initial_plateau_num_warn = 1.5*200, #*min_live_points,
		widen_before_initial_plateau_num_max = 2*200 #*min_live_points
	)
	
	sampler_run_kwargs_defaults.update(sampler_run_kwargs)
	sampler_run_kwargs = sampler_run_kwargs_defaults
	
	
	
	_lgr.debug(f'{reactive_nested_sampler_kwargs=}')
	
	if 'log_dir' not in reactive_nested_sampler_kwargs:
		raise RuntimeError(f'Need to pass "log_dir" as one of "reactive_nested_sampler_kwargs" to tell ultranest where to store results')
	
	
	def fitting_function(params, objective_function, var_param_name_order, const_param_name_order): # should return fitted parameters
		sampler = ultranest.ReactiveNestedSampler(
			var_param_name_order, 
			objective_function,
			params.get_linear_transform_to_domain(var_param_name_order, (0,1)),
			**reactive_nested_sampler_kwargs
		)
		
		final_result = None
		
		for result in sampler.run_iter(**sampler_run_kwargs):
			sampler.print_results()
			sampler.plot()
			final_result = result
	
		return final_result['maximum_likelihood']['point']
		
	return fitting_function

class UltranestResultSet:
	"""
	Defines some routines for interacting with a folder that has ultranest runs in it. Allows association of
	metadata with ultranest runs and some simple plotting routines.
	"""
	
	metadata_file : str = 'result_set_metadata.json'
	
	def __init__(self, result_set_directory : Path | str):
		"""
		Initialises a result set
		
		# ARGUMENTS # 
			result_set_directory : Path | str
				The directory where ultranest should put it's run_N folders.
		"""
		self.directory = Path(result_set_directory)
		self.metadata = dict()
		self.metadata_path = self.directory / self.metadata_file
		if self.metadata_path.exists():
			self.load_metadata()
	
	def __repr__(self):
		return f'UltransetResultSet({self.directory.absolute()})'
	
	def load_metadata(self):
		"""
		Loads metadata for the result set (saved as a json file)
		"""
		with open(self.metadata_path, 'r') as f:
			self.metadata.update(json.load(f))
	
	def save_metadata(self, make_parent_dirs : bool = True):
		"""
		Saves metadata for the result set as a json file.
		
		# ARGUMENTS #
			make_parent_dirs : bool
				If `True` (default) will create parent directory (the result set directory) when saving the metadata
		"""
		if make_parent_dirs:
			self.metadata_path.parent.mkdir(parents=True, exist_ok=True)
		with open(self.metadata_path, 'w') as f:
			json.dump(self.metadata, f)

	def clear_metadata(self):
		"""
		Remove all metadata from this instance. Will not update the saved metadata until `self.save_metadata()` is called.
		"""
		self.metadata = dict()

	def get_result_data_path(self, idx):
		"""
		Ultranest has "result_set_directory/run[INT]" to hold data for each run
		"""
		return self.directory / f'run{idx}'
		
	def get_result_data_from_path(self, result_data_path : Path):
		"""
		Pulls the results of an ultranest run from a "result_set_directory/run[INT]" path.
		"""
		fname = result_data_path / 'info' / 'results.json'
		
		with open(fname, 'r') as f:
			rdata = json.load(f)
		
		return {
			'param_names': rdata['paramnames'],
			'best_point' : rdata['maximum_likelihood']['point'],
			'stats' : rdata['posterior']
		}

	def get_run_indices(self) -> tuple[int,...] | list[int,...]:
		"""
		Gets the indices of all runs in a result set
		"""
		run_indices = []
		for p in self.directory.iterdir():
			if p.is_dir() and p.stem.startswith('run'):
				run_indices.append(int(p.stem[3:]))
		return sorted(run_indices)


	def get_params_vs_run_index(self, metadata_keys_to_include : list [str] = [], sort_by : str | None = None) -> tuple[list[int,...], dict[str,np.array], dict[str,np.array]]:
		"""
		Gets all parameters (and optionally metadata) for each run in the resultset
		
		# ARGUMENTS #
			metadata_keys_to_include : list[str]
				Will include these metadata keys in the results
			sort_by : None | str = None
				Metadata or parameter name to sort by. If `None`, sort by run index
		
		# RETURNS #
			run_indices : list[int,...]
				A list of all run indices found
			metadata_values : dict[str, np.ndarray]
				A dictionary of metadata values for each run
			param_values : dict[str, np.ndarray]
				A dictionary of parameter values for each run
		"""
		run_indices = self.get_run_indices()
		n_runs_indices = len(run_indices)
		
		metadata_values = {}
		param_values = {}
		
		
		# use numpy arrays
		for i, idx in enumerate(run_indices):
			
			for key in metadata_keys_to_include:
				if key not in metadata_values:
					metadata_values[key] = np.full((n_runs_indices,), np.nan)
				metadata_values[key][i] = self.metadata[k][i] 
		
		
			result = self.get_result_data_from_path(self.get_result_data_path(idx))
			for j, pname in enumerate(result['param_names']):
				if pname not in param_values:
					param_values[pname] = np.full((n_runs_indices,), np.nan)
				param_values[pname][i] = result['best_point'][j]
		
		sort_values = None
		if sort_by in metadata_values:
			sort_values = metadata_values[sort_by]
		elif sort_by in param_values:
			sort_values = param_values[sort_by]
		else:
			sort_values = run_indices
		
		sort_indices = np.argsort(sort_values)
		metadata_values = dict((k,v[sort_indices]) for k,v in metadata_values.items())
		param_values = dict((k,v[sort_indices]) for k,v in param_values.items())
		
		_lgr.debug(f'{metadata_values=} {param_values=}')
		
		return run_indices, metadata_values, param_values

	def plot_params_vs_run_index(self, x_var=None, show=False, save=True, **kwargs):
		"""
		x_var : str | None
			Variable to display along x-axis. If `None`, uses run index.
		show : bool = False
			If `True` will display plots
		save : bool = True
			If `True` will save plots to `self.plot_directory`
		kwargs : dict[str,Any]
			Forwarded to `self.get_params_vs_run_index(...)`
		"""
		run_indices, metadata_values, param_values = self.get_params_vs_run_index(**kwargs)
		
		x_label = 'run index' if x_var is None else x_var
		
		x_values = None
		if x_var in metadata_values:
			x_values = metadata_values[x_var]
		elif x_var in param_values:
			x_values = param_values[x_var]
		else:
			x_values = run_indices
			
		
		fname = f'params_vs_{x_label}.png'
		
		f, a = plot_helper.figure_n_subplots(len(param_values)+len(metadata_values))
		#f.tight_layout(pad=16, w_pad=8, h_pad=4)
		f.set_layout_engine('constrained')
		
		f.suptitle(f'params and metadata vs {x_label}')
		for i, pname in enumerate(param_values):
			a[i].set_title(pname)
			a[i].set_xlabel(x_label)
			a[i].set_ylabel(pname)
			a[i].plot(x_values, param_values[pname], 'bo-')
		
		for i, mname in enumerate(metadata_values):
			a[i].set_title(mname)
			a[i].set_xlabel(x_label)
			a[i].set_ylabel(mname)
			a[i].plot(x_values, metadata_values[mname], 'bo-')
		
		
		plot_helper.output(
			show, 
			None if save is None else self.plot_directory / fname
		)
	
	@property
	def plot_directory(self) -> Path:
		"""
		Directory (within `self.directory`) where plots are saved. Defaults to `self.directory / "resultset_plots"
		"""
		plot_dir = self.directory / "resultset_plots"
		plot_dir.mkdir(parents=True, exist_ok=True) # ensure the directory exists if we ask for it
		return plot_dir
	
	def plot_results(self, 
			result_providers : list[Callable[tuple[float,...], np.ndarray]], # List of functions that produce results for run "N" 
			ref_data : np.ndarray, 
			show : bool = False, 
			save : bool = True
		):
		"""
		Plots all results in the resultset
		
		# ARGUMENTS #
			result_providers : list[Callable[tuple[float,...], np.ndarray]]
				A list of callables that accept an ordered tuple of results, much like objective functions for the scipy library.
				Should be ordered corresponding to the index of a run (i.e. 0th entry gives results for lowest run number)
			ref_data : np.ndarray
				An array of data, it should have an entry for each run index (i.e. for run folder "run_128", we use ref_data[128])
			show : bool = False
				If `True` will show the plots as they are made
			save : bool = True
				If `True` will save the results to `self.plot_directory`
		"""
		log_plot_fname_fmt = 'log_result_{idx}.png'
		linear_plot_fname_fmt = 'linear_result_{idx}.png'
		
		run_indices, metadata_values, param_values = self.get_params_vs_run_index()
		
		
		# Loop over run indices and plot all the things we need
		for i, idx in enumerate(run_indices):
			_lgr.debug(f'{i=} {idx=}')
			
			params_for_idx = dict((k,v[i]) for k,v in param_values.items())
			
			result = result_providers[i](params_for_idx)
			
			
			data = ref_data[idx]
			log_data = np.log(data)
			log_data[np.isinf(log_data)] = np.nan
			
			
			
			
			# plot log of result vs reference data
			f, a = plot_helper.figure_n_subplots(4)
			f.set_layout_engine('constrained')
			f.suptitle(f'log results {idx=}\n{params_for_idx}')
			vmin, vmax = np.nanmin(log_data), np.nanmax(log_data)
			
			a[0].set_title(f'log data [{vmin}, {vmax}]')
			a[0].imshow(log_data, vmin=vmin, vmax=vmax)
			
			a[1].set_title(f'log result [{vmin}, {vmax}]')
			a[1].imshow(np.log(result), vmin=vmin, vmax=vmax)
			
			a[2].set_title(f'log residual [{vmin}, {vmax}]')
			a[2].imshow(np.log(data-result), vmin=vmin, vmax=vmax)
			
			log_abs_residual = np.log(np.abs(data-result))
			a[3].set_title(f'log abs residual [{np.nanmin(log_abs_residual)}, {np.nanmax(log_abs_residual)}]')
			a[3].imshow(log_abs_residual)
			
			plot_helper.output(
				show, 
				None if save is None else self.plot_directory / log_plot_fname_fmt.format(idx=idx)
			)
			
			
			# plot result vs reference data
			f, a = plot_helper.figure_n_subplots(4)
			f.set_layout_engine('constrained')
			f.suptitle(f'linear results {idx=}\n{params_for_idx}')
			
			vmin, vmax = np.nanmin(data), np.nanmax(data)
			
			a[0].set_title(f'data [{vmin}, {vmax}]')
			a[0].imshow(data, vmin=vmin, vmax=vmax)
			
			a[1].set_title(f'result [{vmin}, {vmax}]')
			a[1].imshow(result, vmin=vmin, vmax=vmax)
			
			a[2].set_title(f'residual [{vmin}, {vmax}]')
			a[2].imshow(data-result, vmin=vmin, vmax=vmax)
			
			frac_residual = np.abs(data-result)/data
			fr_sorted = np.sort(frac_residual.flatten())
			vmin=fr_sorted[fr_sorted.size//4]
			vmax = fr_sorted[3*fr_sorted.size//4]
			a[3].set_title(f'frac residual [{vmin}, {vmax}]')
			a[3].imshow(frac_residual, vmin=vmin, vmax=vmax)
			
			
			
			plot_helper.output(
				show, 
				None if save is None else self.plot_directory / linear_plot_fname_fmt.format(idx=idx)
			)



"""
Quick tool for spectrally rebinning a FITS file
"""

import sys
from pathlib import Path

from typing import Literal

import numpy as np
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice

import aopp_deconv_tool.numpy_helper.array.grid

import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')




named_spectral_binning_parameters = dict(
	spex = dict(
		bin_step = 1E-9,
		bin_width = 2E-9
	)
)


def plot_rebin(old_bins, old_data, new_bins, new_data, title=None):
	plt.title(title)
	plt.plot(np.sum(old_bins, axis=0)/2, old_data, label='old_data')
	plt.plot(np.sum(new_bins, axis=0)/2, new_data, label='new_data')
	plt.legend()
	plt.show()




def rebin_hdu_over_axis(
		data_hdu, 
		axis, 
		bin_step : float = 1E-9, 
		bin_width : float = 2E-9,
		operation : Literal['sum'] | Literal['mean'] | Literal['mean_err'] = 'mean',
		plot : bool = False
	) -> tuple[np.ndarray, np.ndarray]:
	ax_values = aph.fits.header.get_world_coords_of_axis(data_hdu.header, axis)
	_lgr.debug(f'{ax_values=}')
	
	old_bins = nph.array.grid.edges_from_midpoints(ax_values)
	_lgr.debug(f'{old_bins=}')
	
	new_bins = nph.array.grid.edges_from_bounds(old_bins[0,0], old_bins[-1,-1], bin_step, bin_width)
	_lgr.debug(f'{new_bins=}')

	match operation:
		case 'sum':
			regrid_data, regrid_bin_weights = nph.array.grid.regrid(data_hdu.data, old_bins, new_bins, axis)
			new_data = regrid_data
		case 'mean':
			regrid_data, regrid_bin_weights = nph.array.grid.regrid(data_hdu.data, old_bins, new_bins, axis)
			new_data = (regrid_data.T/regrid_bin_weights).T
		case 'mean_err':
			regrid_data, regrid_bin_weights = nph.array.grid.regrid(data_hdu.data**2, old_bins, new_bins, axis)
			new_data = (regrid_data.T/(regrid_bin_weights**2)).T
		case _:
			raise RuntimeError(f'Unknown binning operation "{operation}"')
	
	if plot:
		plot_rebin(
			old_bins, 
			data_hdu.data[:, data_hdu.data.shape[1]//2, data_hdu.data.shape[2]//2], 
			new_bins, 
			new_data[:, new_data.shape[1]//2, new_data.shape[2]//2]
		)
	
	return new_bins, new_data


def run(
		fits_spec : aph.fits.specifier.FitsSpecifier, 
		output_path : Path | str, 
		bin_step : float = 1E-9, 
		bin_width : float = 2E-9,
		operation : Literal['sum'] | Literal['mean'] | Literal['mean_err'] = 'mean'
	) -> tuple[np.ndarray, np.ndarray]:

	new_data = None
	with fits.open(Path(fits_spec.path)) as data_hdul:
	
		#_lgr.debug(f'{fits_spec.ext=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
	
		axes_ordering =  aph.fits.header.get_axes_ordering(data_hdu.header, fits_spec.axes['SPECTRAL'])
		axis = axes_ordering[0].numpy
	
		new_spec_bins, new_data = rebin_hdu_over_axis(data_hdu, axis, bin_step, bin_width, operation, plot=False)

	
	
		hdr = data_hdu.header
		axis_fits = axes_ordering[0].fits
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'bin_axis' : axis_fits,
			'bin_step' : bin_step,
			'bin_width' : bin_width,
			'bin_operation' : operation
		}
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='spectral_rebin',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
		
		aph.fits.header.set_axes_transform(hdr, 
			axis_fits, 
			'Angstrom', 
			np.mean(new_spec_bins[:,0])/1E-10,
			bin_step/1E-10,
			new_spec_bins.shape[1],
			1
		)

	
	# Save the products to a FITS file
	hdu_rebinned = fits.PrimaryHDU(
		header = hdr,
		data = new_data
	)
	hdul_output = fits.HDUList([
		hdu_rebinned,
	])
	hdul_output.writeto(output_path, overwrite=True)




def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter
	)
	
	parser.add_argument(
		'fits_spec', 
		help = aopp_deconv_tool.text.wrap(
			aph.fits.specifier.get_help(['SPECTRAL']),
			os.get_terminal_size().columns - 50
		)
	)
	parser.add_argument('-o', '--output_path', help='Output fits file path. By default is same as fie `fits_spec` path with "_rebin" appended to the filename')
	
	parser.add_argument('--rebin_operation', choices=['sum', 'mean', 'mean_err'], default='mean', help='Operation to perform when binning.')
	
	rebin_group = parser.add_mutually_exclusive_group(required=False)
	rebin_group.add_argument('--rebin_preset', choices=list(named_spectral_binning_parameters.keys()), default='spex', help='Rebin according to the spectral resolution of the preset')
	rebin_group.add_argument('--rebin_params', nargs=2, type=float, metavar='float', help='bin_step and bin_width for rebinning operation (meters)')
	
	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, ['SPECTRAL'])
	
	if args.rebin_preset is not None:
		for k,v in named_spectral_binning_parameters[args.rebin_preset].items():
			setattr(args, k, v)
	if args.rebin_params is not None:
		setattr(args, 'bin_step', args.rebin_params[0])
		setattr(args, 'bin_width', args.rebin_params[1])
	
	if args.output_path is None:
		args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+'_rebin'+str(Path(args.fits_spec.path).suffix)))
	
	return args


if __name__ == '__main__':
	args = parse_args(sys.argv[1:])
	
	run(args.fits_spec, bin_step=args.bin_step, bin_width=args.bin_width, operation=args.rebin_operation, output_path=args.output_path)
	



"""
Classes for dealing with a beam of light. I.e. a bundle of rays enveloped by 2
straight lines
"""


import dataclasses as dc
import numpy as np




# TODO: Move this to better location
@dc.dataclass(slots=True)
class LightBeam:
	"""
	A beam of light defined by the two lines that limit the extent of the beam,
	see (FIG 1). The beam behaves according to geometric optics. Beam position
	is defined in terms of optica path distance "o". Therefore a beam never goes
	"backwards".
	
	w_a = c_a + o*m_a
	w_b = c_b + o*m_b
	
	w_a - The line that defines the beams "inner" edge, closest to the optical axis
	w_b - The line that defines the beams "outer" edge, farthest from the optical axis
	
	
	
	
	
	FIG 1 ######################################################################
	
	      |......................... w_b - outer edge of light beam
	     /|
	    / |
	   /  |       _________________ w_a - iner edge of light beam
	  /   |      X
	 /    |      X
	*-----|------X----------------- optical axis (o)
	
	             ^
	             Obstruction
	      ^
	      Lens
	^
	Light Source
	
	############################################################################
	
	
	"""
	c_a : float = 0 # y intercept of inner edge
	m_a : float = 0 # gradient of inner edge
	c_b : float = 1 # y intercept of outer edge
	m_b : float = 0 # gradient of outer edge
	o : float = 0 # starting point (on optical axis) of this light beam
	
	def w_a(self, optical_position):
		"""
		The width (distance from the undeviated ray/optical axis) of the inner edge of the beam envelope
		"""
		return self.c_a + self.m_a*(optical_position-self.o)
	
	def w_b(self, optical_position):
		"""
		The width (distance from the undeviated ray/optical axis) of the outer edge of the beam envelope
		"""
		return self.c_b + self.m_b*(optical_position-self.o)
	
	def is_null(self) -> bool:
		"""
		Returns TRUE if the beam is null. I.e. represents no light in a region.
		"""
		return any(np.isnan(x) for x in (self.c_a, self.m_a, self.c_b, self.c_b, self.o))
	
	def __call__(self, optical_position):
		"""
		Calculate the beam envelope at `optical_position`
		"""
		return self.w_a(optical_position), self.w_b(optical_position)
	
	def get_focal_length(self) -> float:
		"""
		For a beam, calculate the focal length that it's angle w.r.t the 
		undeviated ray/optical axis would correspond to.
		"""
		f_a =  self.c_a / self.m_a
		f_b = self.c_b / self.m_b
		assert f_a - f_b < 1E-6, f"Focal lengths of inner and outer edge of beam need to match to within tolerance {f_a=} {f_b=}"
		return f_a


@dc.dataclass(slots=True)
class LightBeamSet:
	"""
	A set of light beams that are present in an optical system. Each beam exists
	between two components (or before the first and after the final components).
	
	Note, a light beam set is defined in optical distance "o", like a beam so the
	members of a beam set are monotonically increasing in "o".
	"""
	light_beams : list[LightBeam] = dc.field(default_factory=list)
	
	def get_optical_position_range(self):
		"""
		Get the range of optical positions covered by the beams in the set
		"""
		vmin = np.inf
		vmax = -np.inf
		for lb in self.light_beams:
			if lb.o < vmin: vmin = lb.o
			if lb.o > vmax: vmax = lb.o
		return vmin, vmax
	
	def get_light_beam_at_position(self, optical_position) -> LightBeam:
		"""
		For some `optical_position`, return the beam in the set that is at that
		position.
		"""
		if optical_position <= self.light_beams[0].o:
			return None
			
		for i, lb in enumerate(self.light_beams):
			if lb.o >= optical_position:
				return self.light_beams[i-1]
		return self.light_beams[-1]
	
	def __getitem__(self, idx):
		"""
		Return the `idx`-th beam from the start of the beam set.
		"""
		return self.light_beams[idx]
	
	def __call__(self, optical_position):
		"""
		Get w_a, w_b for light beam that is part of the set at `optical_position`
		"""
		# Assume that list is ordered by optical distance at this point
		lb = self.get_light_beam_at_position(optical_position)
		if lb is None:
			return np.nan, np.nan
		return lb(optical_position)
 




import numpy as np

from aopp_deconv_tool.geo_array import plot_ga
from aopp_deconv_tool.optics.function import PupilFunction, PointSpreadFunction, OpticalTransferFunction 
from aopp_deconv_tool.optics.geometric.optical_component import OpticalComponentSet

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


def pupil_function_of_optical_component_set(
		shape : tuple[int,...],
		expansion_factor : float,
		supersample_factor : float,
		ocs : OpticalComponentSet,
		scale : tuple[float,...],
	) -> PupilFunction:
	"""
	Get the pupil function of an optical component set
	"""
	if scale is None: 
		scale = ocs.get_pupil_function_scale(expansion_factor)
	return PupilFunction(
		ocs.pupil_function(shape, scale, expansion_factor, supersample_factor),
		tuple(np.linspace(-scale*expansion_factor/2,scale*expansion_factor/2,int(s*expansion_factor*supersample_factor)) for scale, s in zip(scale, shape)),
	)


def optical_transfer_function_of_optical_component_set(
		shape : tuple[int,...],
		expansion_factor : float,
		supersample_factor : float,
		ocs : OpticalComponentSet,
		scale : tuple[float,...],
	) -> OpticalTransferFunction:
	_lgr.debug(f'{shape=} {scale=}')
	pupil_function_axes = tuple(np.fft.fftshift(np.fft.fftfreq(sh, sc/sh)) for sh, sc in zip(shape,scale))
	pupil_function_scale = np.array([x[-1] - x[0] for x in pupil_function_axes])
	_lgr.debug(f'{pupil_function_scale=}')
	
	
	pupil_function = pupil_function_of_optical_component_set(
		shape,
		expansion_factor,
		supersample_factor,
		ocs,
		pupil_function_scale
	)
	# DEBUGGING
	#plot_ga(pupil_function)
	
	# Note PSF for a disk should be an Airy disk, the location of the first minimum, theta ~ 1.22*wavelength/obj_diameter
	# so theta/wavelength ~ 1.22/obj_diameter = 0.1525 for obj_diameter = 8
	# remember that this is approximate, not exact.
	psf = PointSpreadFunction.from_pupil_function(pupil_function)
	otf = OpticalTransferFunction.from_psf(psf)
	otf.data /= np.nansum(otf.data)
	return otf



"""
Helpers for array operations involving indexing
"""
from typing import TypeVar, NewType

import numpy as np


import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'WARNING')


from aopp_deconv_tool.typedef import NumVar, ShapeVar

T = TypeVar('T')

type N = NewType('N', NumVar)
type M = NewType('M', NumVar)

type S[X] = NewType('ShapeS', ShapeVar[X])
type Q[X] = NewType('ShapeQ', ShapeVar[X])
type R[X] = NewType('ShapeR', ShapeVar[X])


def pacman(a : np.ndarray[S[N],T], indices : np.ndarray[[N,M],int]) -> np.ndarray[[N,M],T]:
	"""
	Get values of array `a` at `indices`, out of bounds indices obey pacman physics,
	i.e. topology of a torus. Same as periodic boundaries
	"""
	return a[tuple(i % s for i,s in zip(indices, a.shape))]

def periodic(a : np.ndarray[S[N],T], indices : np.ndarray[[N,M],int]) -> np.ndarray[[N,M],T]:
	"""
	Get values of array `a` at `indices`, out of bounds indices are wrapped periodically
	"""
	return pacman(indices, a)

def const(a : np.ndarray[S[N],T], indices : np.ndarray[[N,M],int], const : T = 0) -> np.ndarray[[N,M],T]:
	"""
	Get values of array `a` at `indices`, out of bounds indices return a constant
	"""
	# get a mask that has the out of bounds elements, set them to a in-bound
	# index to get the values, then set the values from the out of bounds
	# indices to a constant.
	idx_out_of_bounds_mask = np.array([(i < 0) | (i >= s) for i,s in zip(indices, a.shape)],bool)
	indices[idx_out_of_bounds_mask] = 0
	oob_mask = np.array(idx_out_of_bounds_mask.sum(axis=0),bool)
	v = a[(*indices,)]
	v[oob_mask] = const
	return v

def reflect(a : np.ndarray[S[N],T], indices : np.ndarray[[N,M],int]) -> np.ndarray[[N,M],T]:
	"""
	Get values of array `a` at `indices`, out of bounds indices are reflected at the edges.
	"""
	# if i < s, then i_reflect = -i
	# if i >=s, then i_reflect = 2*(s-1) - i
	# Use multiplication to do if statements in arrays
	return a[tuple(i + 2*(s-1)*(i>=s) + i*(-2*(~((0<=i) & (i<s)))) for i,s in zip(indices, a.shape))] 



"""
Uses SSA to get values for interpolation of data
"""

import numpy as np
import scipy as sp
import scipy.stats
import scipy.ndimage
import matplotlib.pyplot as plt


import aopp_deconv_tool.context as context
from aopp_deconv_tool.context.next import Next
from aopp_deconv_tool.stats.empirical import EmpiricalDistribution
from aopp_deconv_tool.scipy_helper.interp import interpolate_at_mask
import aopp_deconv_tool.plot_helper as plot_helper



import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


def ssa_intepolate_at_mask(
		ssa,
		mask,
		start=0, 
		stop=None, 
		value=0.5, # 0.5 seems to be the best value
		show_plots=0,
	):
	"""
	Using SSA for interpolation. E.g. 
		1) pass in a set of pixels
		2) Calculate the 'difference from median of SSA component' score for each pixel
		3) For each pixel, only combine SSA components when the |score| < 'some value'
		This should ensure that 'extreme' values for that pixel are ignored and the 
		reconstructed pixel value is more similar to the surrounding pixels.
	
	This should help when interpolating across a e.g. "hot" pixel as the components
	that contribute to the "hotness" will be ignored and the components that are more
	similar to the surroundings will contribute.
	"""
	# set up arrays and defaults
	stop = ssa.X_ssa.shape[0] if stop is None else stop
	start = 0 if start is None else start
	data_probs = np.zeros((stop-start, *ssa.a.shape),dtype=float)
	
	# choose a pixel->probability function
	# we get a range of (0,1) from "ut.sp.construct_cdf_from()"
	# it gives us the cumulative probability for each pixel
	# This function changes it to 'distance away from median' in the range [-1, 1]
	prob_median_transform_func = lambda x: (2*(x-0.5))
	
	# use the mask to 
	interp_mask_accumulator = np.zeros((np.count_nonzero(mask),), dtype=float)
	px_contrib_mask = np.zeros_like(interp_mask_accumulator, dtype=bool)
	px_contrib_temp = np.zeros_like(interp_mask_accumulator, dtype=float)
	
	# apply pixel->probability function to each SSA component
	for i in range(ssa.X_ssa.shape[0]):
		_lgr.debug(f'i={i}/{ssa.X_ssa.shape[0]}')
		data_distribution = EmpiricalDistribution(ssa.X_ssa[i].ravel())
		
		px_contrib_temp[:] = ssa.X_ssa[i][mask]
		
		_lgr.debug(f'{px_contrib_temp=}')

		if i >= start and i < stop:
			j = i - start

			# Interpolate the SSA components
			#points = np.indices(ssa.X_ssa[i].shape)
			#p_known = points[:,~mask].T
			#p_unknown = points[:,mask].T
			#known_values = ssa.X_ssa[i][~mask]
			#interp_values = sp.interpolate.griddata(p_known, known_values, p_unknown)
			interp_values = interpolate_at_mask(ssa.X_ssa[i], mask, edges='convolution')[mask]
			
			# Get the distance from the median for each pixel
			data_probs[j,...] = prob_median_transform_func(data_distribution.cdf(ssa.X_ssa[j].ravel()).reshape(ssa.a.shape))
			
			# If masked pixel is close to the median, don't replace it. Otherwise, replace with interpolated value
			px_contrib_mask[...] = np.fabs(data_probs[j, mask]) >= value
			_lgr.debug(f'{px_contrib_mask=}')
			
			# Only use the interpolated components when we have extreme values
			px_contrib_temp[px_contrib_mask] = interp_values[px_contrib_mask]
		
		_lgr.debug(f'{px_contrib_temp=}')
		interp_mask_accumulator += px_contrib_temp
		_lgr.debug(f'{interp_mask_accumulator=}')

	result = np.sum(ssa.X_ssa, axis=0)
	result[mask] = interp_mask_accumulator
	
	return result


"""
Routine for parsing a string containing the path, extension, slices, and axes
of a FITS file that we want to operate on.
"""
import os
from collections import namedtuple 

from astropy.io import fits


import aopp_deconv_tool.text as text
import aopp_deconv_tool.cast as cast

import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.slice

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.header

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


FitsSpecifier = namedtuple('FitsSpecifier', ('path', 'ext', 'slices', 'axes'))



AxesInfo = namedtuple('AxesInfo', ('description', 'default_callable'))
axes_type_info={
	"SPECTRAL" : AxesInfo("wavelength/frequency varies along this axis", aph.fits.header.get_spectral_axes),
	"CELESTIAL" : AxesInfo("sky position varies along this axis", aph.fits.header.get_celestial_axes),
	"POLARISATION" : AxesInfo("polarisation varies along this axis, could be linear of circular", aph.fits.header.get_polarisation_axes),
	"TIME" : AxesInfo("time varies along this axis", aph.fits.header.get_time_axes),
}

help_fmt = """\
Format:
	`.../path/to/a.fits{{ext}}[slice0,slice1,...]{{axes_type_1:(ax11,ax12,...),axes_type_2:(ax21,ax22,...)}}
	
	Where:
		.../path/to/a.fits
			A path to a FITS file. Must be present.
		ext : str | int
			Fits extension, either a string or an index. If not present, will assume PRIMARY hdu, (hdu index 0)
		sliceM : str
			Slice of Mth axis in the normal python `start:stop:step` format. If not present will assume slice(None) for an axis.
		axes_type_N : axes_type
			Type of the Nth axes set. Axes types are detailed below. They are used to tell a program what to do with
			an axis when normal FITS methods of description fail (or as a backup). Note, the type (and enclosing curly backets
			`{{}}` can be ommited if a program only accepts one or two axes types. In that case, the specified axes will be
			assumed to be of the first type specified in the documentation, and the remaining axes of the second type (if there
			is a second type).
		axNL : int
			Axes of the extension that are of the specified type. If not present, will assume all axes are of the first type
			specified in the documentation.
	
	Accepted axes_type:
		{axes_types}
		
	
	Examples:
		~/home/datasets/MUSE/neptune_obs_1.fits{{DATA}}[100:200,:,:](1,2)
			Selects the "DATA" extension, slices the 0th axis from 100->200 leaving the others untouched, and signals that axes 1 and 2 are important e.g. they are the RA-DEC axes.
		~/home/datasets/MUSE/neptune_obs_1.fits{{DATA}}[100:200](1,2)
			Does the same thing as above, but omits un-needed slice specifiers.
		~/home/datasets/MUSE/neptune_obs_1.fits{{DATA}}[100:200]{{CELESTIAL:(1,2)}}
			Again, same as above, but adds explicit axes type.
"""

def get_help(axes_types : list[str]):
	"""
	Generates the help string for a list of axes_types.
	"""
	for x in axes_types:
		if x not in axes_type_info:
			raise RuntimeError(f"axes_type '{x}' is not one of the known axes_types {list(axes_type_info.keys())}")
	
	return help_fmt.format(axes_types='\n\t\t'.join([k+'\n\t\t\t'+axes_type_info[k].description for k in axes_types]))



def parse_axes_type_list(axes_type_list : str, axes_types: list[str] | tuple[str]):
	"""
	Parses the axes type list at the end of the specifier string.
	
	Arguments:
		axes_type_list : str
			"axes_type_1:(ax11,ax22,...),axes_type_2:(ax21,ax22,...)" or "(ax11,ax22)"
		axes_types : list[str] | tuple[str]
			List of acceptable axes types
	"""
	
	# first split on commas
	axes = {}
	#_lgr.debug(f"{text.split_around_brackets(axes_type_list)=}")
	for i, axes_type_str in enumerate(text.split_around_brackets(axes_type_list)):
		# axes_type_str = "axes_type_1:(ax11,ax12,...)" or "(ax11,ax12,...)"
		#_lgr.debug(f'{axes_type_str=}')
		n_colon = axes_type_str.count(':')
		if n_colon == 0:
			axtype, axtuple_str = (axes_types[i], axes_type_str)
		elif n_colon == 1:
			axtype, axtuple_str = axes_type_str.split(':')
		else:
			raise RuntimeError("axes_type_str '{axes_type_str}' has more than one colon, it is malformed.")
		
		# now parse axtuple_str
		assert axtuple_str[0] == '(' and axtuple_str[-1] ==')', f"axes tuple '{axtuple_str}' must be enclosed in brackets"
		
		axtuple = text.to_tuple(axtuple_str[1:-1], int)
		axes[axtype] = axtuple
	
	return(axes)
		
	

def parse(specifier : str, axes_types : list[str]):
	f"""
	Parses a string that specifies a fits file, extension, and slices.
	
	# ARGUMENTS #
		specifier : str
			A FITS specifier string, has the following format:
				{text.indent(get_help(axes_types), 4)}
		axes_types : list[str]
			A list of axes types that we expect. Should be a list of entries from `axes_type_info.keys()`
	
	# RETURNS #
		FitsSpecifier(path, ext, slices, axes)
			
	"""
	help_string = get_help(axes_types)
	
	# Go through cases from end to start
	
	try:
		# get axes information
		j = len(specifier) - 1
		axes_type_list = None
		if specifier[j-1:] == ')}':
			# we have axes_types in the specifier.
			i = specifier.rfind('{',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '{'")
			axes_type_list = specifier[i+1:j]
			specifier = specifier[:i]
			j = i-1
		elif specifier[j] == ')':
			i = specifier.rfind('(',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '('")
			axes_type_list = specifier[i:j+1]
			specifier = specifier[:i]
			j = i-1
		#_lgr.debug(f'{axes_type_list=}')	
		if axes_type_list is not None:
			axes = parse_axes_type_list(axes_type_list, axes_types)
		else:
			axes = None
		#_lgr.debug(f'{axes=}')
			
			
		
		# get slice information
		if specifier[j] == ']':
			# we have slice information
			i = specifier.rfind('[',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '['")
			slice_tuple_str = specifier[i+1:j] # exclude "[" and "]" to give "a:b:c,d:e:f,..."
			slices = nph.slice.from_string(slice_tuple_str)
			specifier = specifier[:i]
			j = i-1
		else:
			slices=None
		
		
		# get extension
		if specifier[j] == '}':
			i = specifier.rfind('{',0,j)
			if i == -1:
				raise RuntimeError("FITS specifier is malformed, missing '{'")
			ext_str = specifier[i+1:j]
			ext = cast.to_any(ext_str, (int, str))
			specifier = specifier[:i]
			j = i-1
		else:
			ext = "DATA"
		
	except Exception as e:
		e.add_note(help_string)
		raise
	
	# get path
	path = specifier
	
	if not os.path.isfile(path):
		raise FileNotFoundError(f'file "{path}" not found')

	if slices is None or axes is None:
		try:
			hdr = fits.getheader(path,ext=ext if type(ext) is int else (ext,1))
		except Exception as e:
			pass
		else:
			if slices is None:
				slices = tuple(slice(None) for _ in range(hdr['NAXIS']))
			if axes is None:
				axes = {}
				for ax_name in axes_types:
					axes[ax_name] = axes_type_info[ax_name].default_callable(hdr)


	return FitsSpecifier(path, ext, slices, axes)



"""
Quick tool for normalising a PSF in a FITS file
"""


import sys
from pathlib import Path

from typing import Literal

import numpy as np
import scipy as sp
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice

import aopp_deconv_tool.psf_data_ops as psf_data_ops


import matplotlib.pyplot as plt

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


def run(
		fits_spec,
		output_path,
		threshold : float = 1E-2,
		n_largest_regions : None | int = 1,
		background_threshold : float = 1E-3,
		background_noise_model : str = 'gennorm',
		n_sigma : float = 5,
		trim_to_shape : None | tuple[int,...] = None
	):
	
	ACCEPTABLE_TRIM_LOSS_FACTOR = 1E-2
	axes = fits_spec.axes['CELESTIAL']
	
	with fits.open(Path(fits_spec.path)) as data_hdul:
		
		
		_lgr.debug(f'{fits_spec.path=} {fits_spec.ext=} {fits_spec.slices=} {fits_spec.axes=}')
		#raise RuntimeError(f'DEBUGGING')
	
		data_hdu = data_hdul[fits_spec.ext]
		data = data_hdu.data[fits_spec.slices]
		hdr = data_hdu.header
		
		
		_lgr.info('Ensure data is of odd shape')
		data = nph.array.ensure_odd_shape(data, axes)
		for ax in axes:
			aph.fits.header.set_axes_transform(hdr, ax, n_values = data.shape[ax])
		
		_lgr.info('Remove any outliers')
		outlier_mask = psf_data_ops.get_outlier_mask(data, axes, n_sigma)
		data[outlier_mask] = np.nan
		
		if background_noise_model is not None:
			_lgr.info('Remove background offset')
			background_mask = ~psf_data_ops.get_roi_mask(data, axes, background_threshold, 1)
			noise_model_offsets, noise_model_parameters, noise_model_at_values, noise_model_cdf, noise_model_cdf_residual = psf_data_ops.remove_offset(data, axes, background_mask, background_noise_model)
		
		_lgr.info('Get offsets to center around center of mass')
		roi_mask = psf_data_ops.get_roi_mask(data, axes, threshold, n_largest_regions)
		com_offsets = psf_data_ops.get_center_of_mass_offsets(data, axes, roi_mask)
	
		_lgr.info('Recenter everything for easy comparison')
		normalised_data = psf_data_ops.apply_offsets(data, axes, com_offsets)
		roi_mask = psf_data_ops.apply_offsets(roi_mask, axes, com_offsets)
		background_mask = psf_data_ops.apply_offsets(background_mask, axes, com_offsets)
		outlier_mask = psf_data_ops.apply_offsets(outlier_mask, axes, com_offsets)
		
		
		if trim_to_shape is not None:
			_lgr.info('Trim everything around the center pixel')
			before_trim_sum = np.nansum(normalised_data)
			
			normalised_data = psf_data_ops.trim_around_center(normalised_data, axes, trim_to_shape)
			roi_mask = psf_data_ops.trim_around_center(roi_mask, axes, trim_to_shape)
			outlier_mask = psf_data_ops.trim_around_center(outlier_mask, axes, trim_to_shape)
			
			after_trim_sum = np.nansum(normalised_data)
			trim_loss_factor = 1 - after_trim_sum/before_trim_sum
			_lgr.debug(f'{trim_loss_factor=}')
			if trim_loss_factor > ACCEPTABLE_TRIM_LOSS_FACTOR:
				_lgr.warn(f'After trimming around center to shape {trim_to_shape}, have lost {trim_loss_factor} of original signal. This is above acceptable limit of {ACCEPTABLE_TRIM_LOSS_FACTOR}')
			
			
		
		_lgr.info('Normalise to unit sum')
		original_sum = np.nansum(normalised_data, axis=axes)
		with nph.axes.to_start(normalised_data, axes) as (gdata, gaxes):
			gdata /= original_sum
		
		param_dict = {
			'original_file' : Path(fits_spec.path).name, # record the file we used
			'roi_mask.threshold' : threshold,
			'roi_mask.n_largest_regions' : n_largest_regions,
			'background_mask.threshold' : background_threshold,
			'outlier_mask.n_sigma' : n_sigma,
			'trim_to_shape' : trim_to_shape,
		}
		
		hdr.update(aph.fits.header.DictReader(
			param_dict,
			prefix='psf_normalise',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
		

	
	_lgr.info('save the products to a FITS file')
	hdus = []
	
	hdus.append(fits.PrimaryHDU(
		header = hdr,
		data = normalised_data
	))
	hdus.append(fits.ImageHDU(
		header = hdr,
		data = roi_mask.astype(int),
		name = 'ROI_MASK'
	))
	hdus.append(fits.ImageHDU(
		header = hdr,
		data = background_mask.astype(int),
		name = 'BACKGROUND_MASK'
	))
	hdus.append(fits.ImageHDU(
		header = hdr,
		data = outlier_mask.astype(int),
		name = 'OUTLIER_MASK'
	))
	hdus.append(fits.BinTableHDU.from_columns(
		columns = [
			fits.Column(name='original_total_sum', format='D', array=original_sum),
		],
		name = 'ORIG_SUM',
		header = None,
	))
	
	if background_noise_model is not None:
		hdus.append(fits.BinTableHDU.from_columns(
			columns = [
				fits.Column(name='noise_model_offsets', format='D', array=noise_model_offsets), 
			] + [
				fits.Column(name=f'noise_model_parameter_{k}', format='D', array=[x[k] for x in noise_model_parameters]) for k in noise_model_parameters[0].keys()
			],
			name = 'NOISE_MODEL_PARAMS',
			header = None,
		))
		
		hdus.append(fits.ImageHDU(
			header=None,
			name='NOISE_MODEL_CDF_VALUES',
			data = np.array(noise_model_at_values)
		))
		
		hdus.append(fits.ImageHDU(
			header=None,
			name='NOISE_MODEL_CDF',
			data = np.array(noise_model_cdf)
		))
		
		
		hdus.append(fits.ImageHDU(
			header=None,
			name='NOISE_MODEL_CDF_RESIDUAL',
			data = np.array(noise_model_cdf_residual)
		))
		
	
	hdul_output = fits.HDUList(hdus)
	hdul_output.writeto(output_path, overwrite=True)


def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_normalised'
	DESIRED_FITS_AXES = 'CELESTIAL'
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter
	)
	
	parser.add_argument(
		'fits_spec', 
		help = aopp_deconv_tool.text.wrap(
			aph.fits.specifier.get_help([DESIRED_FITS_AXES]).replace('\t', '    '),
			os.get_terminal_size().columns - 30
		)
	)
	parser.add_argument('-o', '--output_path', help=f'Output fits file path. By default is same as fie `fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	
	parser.add_argument('--threshold', type=float, default=1E-2, help='When finding region of interest, only values larger than this fraction of the maximum value are included.')
	parser.add_argument('--n_largest_regions', type=int, default=1, help="""
		When finding region of interest, if using a threshold will only the n_largest_regions in the calculation.
		A region is defined as a contiguous area where value >= `threshold` along `axes`. I.e. in a 3D cube, if
		we recenter about the COM on the sky (CELESTIAL) axes the regions will be calculated on the sky, not in
		the spectral axis (for example)."""
	)
	parser.add_argument('--background_threshold', type=float, default=1E-3, help='Exclude the largest connected region with values larger than this fraction of the maximum value when finding the background')
	parser.add_argument('--background_noise_model', type=str, default='gennorm', 
		choices=['norm', 'gennorm', 'none'],
		help='Model of background noise to use when removing offset. "none" will not mean offset is not calculated or removed'
	)
	
	parser.add_argument('--n_sigma', type=float, default=5, help='When finding the outlier mask, the number of standard deviations away from the mean a pixel must be to be considered an outlier`')
	parser.add_argument('--trim_to_shape', type=int, nargs=2, default=None, help='After centering etc. will trim data to this shape around the center pixel. Used to reduce data volume for faster processing.')

	args = parser.parse_args(argv)
	
	args.fits_spec = aph.fits.specifier.parse(args.fits_spec, [DESIRED_FITS_AXES])
	
	if args.output_path is None:
		args.output_path =  (Path(args.fits_spec.path).parent / (str(Path(args.fits_spec.path).stem)+DEFAULT_OUTPUT_TAG+str(Path(args.fits_spec.path).suffix)))
	if args.background_noise_model == 'none':
		args.background_noise_model = None
	
	
	return args


if __name__ == '__main__':
	args = parse_args(sys.argv[1:])
	
	run(
		args.fits_spec, 
		args.output_path,
		threshold = args.threshold,
		n_largest_regions = args.n_largest_regions,
		background_threshold = args.background_threshold,
		background_noise_model = args.background_noise_model,
		n_sigma = args.n_sigma,
		trim_to_shape = args.trim_to_shape,
	)
	


"""
Helper routines that calculate values or give information about arrays
"""
from typing import TypeVar, Literal, NewType

import numpy as np

from aopp_deconv_tool.typedef import NumVar, ShapeVar

T = TypeVar('T')

type N = NewType('N', NumVar)
type M = NewType('M', NumVar)

type S[X] = NewType('ShapeS', ShapeVar[X])
type Q[X] = NewType('ShapeQ', ShapeVar[X])
type R[X] = NewType('ShapeR', ShapeVar[X])




def get_center_offset_brightest_pixel(a : np.ndarray[S[N],T]) -> np.ndarray[[N],T]:
	"""
	Get the offset from the center of array `a` of the brightest element of `a`
	"""
	if np.all(np.isnan(a)):
		return(np.zeros(a.ndim))
	offset = np.array([s//2 for s in a.shape]) - np.unravel_index(np.nanargmax(a), a.shape)
	return(offset)


def apply_offset(a : np.ndarray[S[N],T], offset : np.ndarray[[N],int]) -> np.ndarray[S[N],T]:
	"""
	Apply `offset` to array `a`, shifting the elements with periodic boundary conditions
	"""
	return	np.roll(a, offset, tuple(range(a.ndim)))


def ensure_odd_shape(a : np.ndarray[S[N],T], axes : tuple[int,...] | None = None) -> np.ndarray[Q[N],T]:
	"""
	Get a slice of array `a` such that each axis has an odd number of entries.
	The resulting slice is always the same size or smaller than `a`
	"""
	if axes is None:
		axes = tuple(range(a.ndim))
	slices = tuple(slice(s-1+s%2) if i in axes else slice(None) for i, s in enumerate(a.shape))
	return a[slices]


def offsets_from_point(shape : S[N], point : np.ndarray[[N],T] = None, scale : S[N] = None) -> np.ndarray[S[N],T]:
	"""
	For an array of shape `shape` get the offsets from a specific point in array coordinates.
	"""
	if point is None:
		point = np.array([s//2 for s in shape]) # center
	if scale is None:
		scale = np.ones_like(point, dtype=float)
	else:
		# -ve values are cell sizes, +ve are array axis sizes.
		scale = np.array([float(x)/s if x > 0 else -x for x,s in zip(scale, shape)], dtype=float) 
	#print(f'{scale=}')
	return np.moveaxis((np.moveaxis(np.indices(shape),0,-1) - point)*scale,-1,0)

def indices_of_mask(m : np.ndarray[S[N],bool]) -> tuple[np.ndarray[[N],int],...]:
	"""
	Get the indices that are masked by mask `m`
	"""
	return np.where(m)

def manhattan_distance_mask(ii : np.ndarray[[*S[N],N,*Q[M]],T], dist : int | tuple[int,...] = 1, index_axis : int = 0) -> np.ndarray[[*S[N],*Q[M]],T]:
	"""
	Calculate a mask that is the manhattan distance `dist` from the array `ii` holding coord information, where the axis `index_axis`
	is has the `N` coordinates.
	"""
	if index_axis != 0:
		np.moveaxis(ii,index_axis,0)
		
	m = np.ones(ii.shape[1:],bool)
	
	if type(dist) is int:
		dist = tuple([dist]*m.ndim)
	elif len(dist) < m.ndim:
		dist = tuple(*dist,*([0]*(m.ndim-len(dist))))
	
	for b, d in zip(ii,dist):
		m &= np.abs(b) <= d
	m &= (np.abs(ii).sum(axis=0) <= np.max(dist))
	return m

def offsets_manhattan_distance(dist : int | tuple[int,...] = 1, ndim : Literal[N] = 3) -> np.ndarray[[N,M],T]:
	"""
	Get the offsets required to visit each element of an array of dimension `ndim` that is `dist` manhattan distance away or less from a point.
	"""
	if type(dist) is int:
		dist = tuple([dist]*ndim)
	elif len(dist) < ndim:
		dist = (*dist,*([0]*(ndim-len(dist))))
	shape = tuple(2*d + 1 for d in dist)
	return offsets_from_center_of_mask(manhattan_distance_mask(offsets_from_point(shape), dist))

def manhattan_distance(shape : S[N], point : np.ndarray[[N],T] = None) -> np.ndarray[S[N],T]:
	"""
	Generate an array of `shape` where each element is the manhattan distances from `point`
	"""
	if point is None:
		point = tuple(s//2 for s in shape)
	return np.abs(np.moveaxis(np.indices(shape),0,-1) - np.array(point)).sum(axis=-1)

def offsets_from_center_of_mask(mask : np.ndarray[S[N],bool]) -> np.ndarray[[N,*S[N]],int]:
	"""
	Get the offsets to all masked elements of `mask` from the center of `mask`.
	"""
	assert all([s%2 == 1 for s in mask.shape]), "mask must have an odd shape to have a unique center"
	return np.argwhere(mask) - np.array([s//2 for s in mask.shape])







"""
Script that deconvolves the first argument with the second argument

Example invocation: 
	`python -m aopp_deconv_tool.deconvolve './example_data/test_rebin.fits{DATA}[10:12]{CELESTIAL:(1,2)}' './example_data/fit_example_psf_000.fits[10:12]{CELESTIAL:(1,2)}'`
"""

import sys
from pathlib import Path
from typing import Literal

import numpy as np
from astropy.io import fits

import aopp_deconv_tool.astropy_helper as aph
import aopp_deconv_tool.astropy_helper.fits.specifier
import aopp_deconv_tool.astropy_helper.fits.header
import aopp_deconv_tool.numpy_helper as nph
import aopp_deconv_tool.numpy_helper.axes
import aopp_deconv_tool.numpy_helper.slice
import aopp_deconv_tool.psf_data_ops as psf_data_ops

from aopp_deconv_tool.algorithm.deconv.clean_modified import CleanModified
from aopp_deconv_tool.algorithm.deconv.lucy_richardson import LucyRichardson

import matplotlib as mpl
import matplotlib.pyplot as plt
import copy
import aopp_deconv_tool.plot_helper as plot_helper
from aopp_deconv_tool.plot_helper.base import AxisDataMapping
from aopp_deconv_tool.plot_helper.plotters import PlotSet, Histogram, VerticalLine, Image, IterativeLineGraph, HorizontalLine

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


def create_plot_set(deconvolver, cadence = 10):
	fig, axes = plot_helper.figure_n_subplots(8)
	axes_iter = iter(axes)
	a7_2 = axes[7].twinx()
	
	try:
		cmap = mpl.cm.get_cmap('bwr_oob')
	except ValueError:
		cmap = copy.copy(mpl.cm.get_cmap('bwr'))
		cmap.set_over('magenta')
		cmap.set_under('green')
		cmap.set_bad('black')
		mpl.cm.register_cmap(name='bwr_oob', cmap=cmap)
	#mpl.rcParams['image.cmap'] = 'user_cmap'
	
	plot_set = PlotSet(
		fig,
		'clean modified step={self.n_frames}',
		cadence=cadence,
		plots = [	
			Histogram(
				'residual', 
				static_frame=False,
				axis_data_mappings = (AxisDataMapping('value','bins',limit_getter=plot_helper.lim), AxisDataMapping('count','_hist',limit_getter=plot_helper.LimRememberExtremes()))
			).attach(next(axes_iter), deconvolver, lambda x: x._residual),
		 	
			VerticalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'color':'red'}
			).attach(axes[0], deconvolver, lambda x: x._pixel_threshold),
			
			Image(
		 		'residual'
		 	).attach(next(axes_iter), deconvolver, lambda x: x._residual),
			
			Image(
		 		'current cleaned'
			).attach(next(axes_iter), deconvolver, lambda x: x._current_cleaned),
			
			Image(
		 		'components'
			).attach(next(axes_iter), deconvolver, lambda x: x._components),
			
			Image(
		 		'selected pixels'
			).attach(next(axes_iter), deconvolver, lambda x: x._selected_px),
			
			Image(
		 		'pixel choice metric',
		 		axis_data_mappings = (AxisDataMapping('x',None), AxisDataMapping('y',None), AxisDataMapping('brightness', '_z_data', plot_helper.LimSymAroundValue(0))),
		 		plt_kwargs={'cmap':'bwr_oob'}
			).attach(next(axes_iter), deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			Histogram(
				'pixel choice metric', 
				static_frame=False,
			).attach(next(axes_iter), deconvolver, lambda x: x._px_choice_img_ptr.val),
			
			IterativeLineGraph(
				'metrics',
				datasource_name='fabs',
				axis_labels = (None, 'fabs value (blue)'),
				static_frame=False,
				plt_kwargs = {},
				ax_funcs=[lambda ax: ax.set_yscale('log')]
			).attach(next(axes_iter), deconvolver, lambda x: np.fabs(np.nanmax(x._residual))),
			
			HorizontalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'linestyle':'--'}
			).attach(axes[7], deconvolver, lambda x: x._fabs_threshold),
			
			IterativeLineGraph(
				'metrics',
				datasource_name='rms',
				axis_labels = (None,'rms value (red)'),
				static_frame=False,
				plt_kwargs={'color':'red'},
				ax_funcs=[lambda ax: ax.set_yscale('log')]
			).attach(a7_2, deconvolver, lambda x: np.sqrt(np.nansum(x._residual**2)/x._residual.size)),
			
			HorizontalLine(
				None, 
				static_frame=False, 
				plt_kwargs={'color':'red', 'linestyle':'--'}
			).attach(a7_2, deconvolver, lambda x: x._rms_threshold),
		]
	)
	return plot_set

def run(
		obs_fits_spec : aph.fits.specifier.FitsSpecifier,
		psf_fits_spec : aph.fits.specifier.FitsSpecifier,
		output_path : str | Path = './deconv.fits',
		deconv_class : Literal[CleanModified] | Literal[LucyRichardson] = CleanModified,
		plot : bool = True,
		deconv_args : list[str,...] = []
	):
	"""
	Given a FitsSpecifier for an observation and a PSF, an output path, and a class that performs deconvolution,
	deconvolves the observation using the PSF.
	
	# ARGUMENTS #
		obs_fits_spec : aph.fits.specifier.FitsSpecifier
			FITS file specifier for observation data, format is PATH{EXT}[SLICE](AXES).
			Where:
				PATH : str
					The path to the FITS file
				EXT : str | int
					The name or number of the FITS extension (defaults to PRIMARY)
				SLICE : "python slice format" (i.e. [1:5, 5:10:2])
					Slice of the FITS extension data to use (defaults to all data)
				AXES : tuple[int,...]
					Axes of the FITS extension that are "spatial" or "celestial" (i.e. RA, DEC),
					by default will try to infer them from the FITS extension header.
		psf_fits_spec : aph.fits.specifier.FitsSpecifier
			FITS file specifier for PSF data, format is same as above
		output_path : str = './deconv.fits'
			Path to output deconvolution to.
		deconv_class : Type
			Class to use for deconvolving, defaults to CleanModified
		plot : bool = True
			If `True` will plot the deconvolution progress
	"""
	deconv_params = parse_deconv_args(deconv_class, deconv_args)
	_lgr.debug(f'{deconv_params=}')
	deconvolver = deconv_class(**deconv_params)

	# Open the fits files
	with fits.open(Path(obs_fits_spec.path)) as obs_hdul, fits.open(Path(psf_fits_spec.path)) as psf_hdul:
		
		# pull out the data we want
		obs_data = obs_hdul[obs_fits_spec.ext].data
		psf_data = psf_hdul[psf_fits_spec.ext].data
		
		# Create holders for deconvolution products
		deconv_components = np.full_like(obs_data, np.nan)
		deconv_residual = np.full_like(obs_data, np.nan)
		
		# Loop over the index range specified by `obs_fits_spec` and `psf_fits_spec`
		for obs_idx, psf_idx in zip(
			nph.slice.iter_indices(obs_data, obs_fits_spec.slices, obs_fits_spec.axes['CELESTIAL']),
			nph.slice.iter_indices(psf_data, psf_fits_spec.slices, psf_fits_spec.axes['CELESTIAL'])
		):
		
		
			# Set up plotting if we want it
			if plot:
				#plt.figure()
				#plt.imshow(obs_data[obs_idx])
				#plt.figure()
				#plt.imshow(psf_data[psf_idx])
				#plt.show()
				plt.close('all')
				plot_set = create_plot_set(deconvolver)
				deconvolver.post_iter_hooks = []
				deconvolver.post_iter_hooks.append(lambda *a, **k: plot_set.update())
				plot_set.show()
			
			# Ensure that we actually have data in this part of the cube
			if np.all(np.isnan(obs_data[obs_idx])) or np.all(np.isnan(psf_data[psf_idx])):
				_lgr.warn('All NAN obs or psf layer detected. Skipping...')
			
			# perform any normalisation and processing
			normed_psf = psf_data_ops.normalise(np.nan_to_num(psf_data[psf_idx]))
			processed_obs = np.nan_to_num(obs_data[obs_idx])
			
			# Store the deconvolution products in the arrays we created earlier
			deconv_components[obs_idx], deconv_residual[obs_idx], deconv_iters = deconvolver(processed_obs, normed_psf)
			
		
		# Save the parameters we used. NOTE: we are only saving the LAST set of parameters as they are all the same
		# in this case. However, if they vary with index they should be recorded with index as well.
		deconv_params = deconvolver.get_parameters()
		
		# Make sure we get all the observaiton header data as well as the deconvolution parameters
		hdr = obs_hdul[obs_fits_spec.ext].header
		hdr.update(aph.fits.header.DictReader(
			{
				'obs_file' : obs_fits_spec.path,
				'psf_file' : psf_fits_spec.path, # record the PSF file we used
				**deconv_params # record the deconvolution parameters we used
			},
			prefix='deconv',
			pkey_count_start=aph.fits.header.DictReader.find_max_pkey_n(hdr)
		))
	
	# Save the deconvolution products to a FITS file
	hdu_components = fits.PrimaryHDU(
		header = hdr,
		data = deconv_components
	)
	hdu_residual = fits.ImageHDU(
		header = hdr,
		data = deconv_residual,
		name = 'RESIDUAL'
	)
	hdul_output = fits.HDUList([
		hdu_components,
		hdu_residual
	])
	hdul_output.writeto(output_path, overwrite=True)


def parse_deconv_args(deconv_class, argv):
	import argparse
	import dataclasses as dc
	import re
	
	# Use this to grab only the first part of the docstring as that should be a short
	# description of the class
	re_empty_line = re.compile(r'^\s*$\s*', flags=re.MULTILINE)
	
	parser = argparse.ArgumentParser(
		description=re_empty_line.split(deconv_class.__doc__,2)[1], 
		formatter_class=argparse.RawTextHelpFormatter,
		add_help=False
	)
	def on_parser_error(err_str):
		print(err_str)
		parser.print_help()
		sys.exit(1)
	
	parser.error = on_parser_error
	
	for field in dc.fields(deconv_class):
		if field.init != True:
			continue
		field_default = field.default if field.default != dc.MISSING else (field.default_factory() if field.default_factory != dc.MISSING else None)
		parser.add_argument(
			'--'+field.name, 
			type=field.type, 
			default= field_default,
			help=field.metadata.get('description', 'DESCRIPTION NOT FOUND') + f' (default = {field_default})',
			metavar=str(field.type)[8:-2]
		)
	
	deconv_args = parser.parse_args(argv)
	
	return vars(deconv_args)
	
	

def parse_args(argv):
	import os
	import aopp_deconv_tool.text
	import argparse
	
	DEFAULT_OUTPUT_TAG = '_deconv'
	DESIRED_FITS_AXES = ['CELESTIAL']
	
	parser = argparse.ArgumentParser(
		description=__doc__, 
		formatter_class=argparse.RawTextHelpFormatter
	)
	
	parser.add_argument(
		'obs_fits_spec', 
		help = aopp_deconv_tool.text.wrap(
			aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
			os.get_terminal_size().columns - 30
		)
	)
	
	parser.add_argument(
		'psf_fits_spec', 
		help = aopp_deconv_tool.text.wrap(
			aph.fits.specifier.get_help(DESIRED_FITS_AXES).replace('\t', '    '),
			os.get_terminal_size().columns - 30
		)
	)
	
	parser.add_argument('-o', '--output_path', help=f'Output fits file path. By default is same as the `fits_spec` path with "{DEFAULT_OUTPUT_TAG}" appended to the filename')
	parser.add_argument('--plot', action='store_true', default=False, help='If present will show progress plots of the deconvolution')
	
	args, deconv_args = parser.parse_known_args(argv)
	
	args.obs_fits_spec = aph.fits.specifier.parse(args.obs_fits_spec, DESIRED_FITS_AXES)
	args.psf_fits_spec = aph.fits.specifier.parse(args.psf_fits_spec, DESIRED_FITS_AXES)
	
	if args.output_path is None:
		args.output_path =  (Path(args.obs_fits_spec.path).parent / (str(Path(args.obs_fits_spec.path).stem)+DEFAULT_OUTPUT_TAG+str(Path(args.obs_fits_spec.path).suffix)))
	
	
	return args, deconv_args


if __name__ == '__main__':

	args, deconv_args = parse_args(sys.argv[1:])
	

	
	run(
		args.obs_fits_spec, 
		args.psf_fits_spec, 
		output_path = args.output_path, 
		plot = args.plot,
		deconv_args = deconv_args,
	)
	



"""
Classes and routines for wrangling PSF model into a format used by various 3rd party optimisation tools.
"""
from typing import Callable, ParamSpec, Any
import dataclasses as dc
import inspect

import numpy as np

from aopp_deconv_tool.data_structures import BiDirectionalMap

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


P = ParamSpec('P')


def linear_transform_factory(
		in_domain : tuple[float,float], 
		out_domain : tuple[float,float]
	) -> Callable[[float],float]:
	"""
	Creates a function that linearly transforms an input variable from `in_domain` to `out_domain`
	"""
	def linear_transform(x):
		x_dash = (x - in_domain[0])/(in_domain[1]-in_domain[0])
		return (x_dash * (out_domain[1]-out_domain[0])) + out_domain[0]
	return linear_transform




@dc.dataclass(slots=True)
class PriorParam:
	"""
	Class that holds parameter information.
	
	name : str
		String used to identify the parameter
	domain : tuple[float,float]
		Range of possible values the parameter can take
	is_const : bool
		Flag that signals if this parameter is a constant, constant parameters will use `const_value`
	const_value : float
		A value to use when const, and for example plots
	"""
	name : str
	domain : tuple[float,float]
	is_const : bool
	const_value : float
	
	def linear_transform_to_domain(self, in_domain=(0,1)):
		"""
		Returns a function that linearly transforms an `in_domain` to the parameter's domain
		"""
		return linear_transform_factory(in_domain, self.domain)
	
	def linear_transform_from_domain(self, out_domain=(0,1)):
		"""
		Returns a function that linearly transforms the parameter's domain to an `out_domain`
		"""
		return linear_transform_factory(self.domain, out_domain)
	
	def __repr__(self):
		return f'PriorParam(name={self.name}, domain={self.domain}, is_const={self.is_const}, const_value={self.const_value})'

	def to_dict(self):
		return {'name':self.name, 'domain':self.domain, 'is_const':self.is_const, 'const_value':self.const_value}

class PriorParamSet:
	"""
	A collection of PriorParam instances, used to define the fitting region for a fitting function.
	"""
	def __init__(self, *prior_params):
		self.prior_params = list(prior_params)
		self.param_name_index_map = BiDirectionalMap()
		self.all_param_index_to_variable_param_index_map = BiDirectionalMap()
		self.all_param_index_to_constant_param_index_map = BiDirectionalMap()
		
		i=0
		j=0
		for k, p in enumerate(self.prior_params):
			self.param_name_index_map[p.name] = k
			if not p.is_const:
				self.all_param_index_to_variable_param_index_map[k] = i
				i+=1
			else:
				self.all_param_index_to_constant_param_index_map[k] = j
				j+=1
		return
	
	def __repr__(self):
		return repr(self.prior_params)
	
	def __len__(self):
		return len(self.prior_params)
	
	def get_linear_transform_to_domain(self, param_names : tuple[str] | list[str], in_domain : tuple[float,float] | list[float,float] | np.ndarray[[2,'N'],float]):
		"""
		Gets a linear transform from some input domain to some output domain for a set of parameter names
		"""
		n = len(param_names)
		if type(in_domain) == tuple or type(in_domain)==list:
			in_domain = np.array([in_domain,]*n).T
		if in_domain.shape != (2,n):
			raise RuntimeError(f"Input domain must be either a tuple, or a numpy array with shape (2,{n})")
			
		out_domain = np.array(tuple(self[param_name].domain for param_name in param_names)).T
		return linear_transform_factory(in_domain, out_domain)
	
	@property
	def all_params(self):
		"""
		The names of all paramters
		"""
		return (*self.variable_params, *self.constant_params)
	
	@property
	def variable_params(self):
		"""
		The names of the variable parameters
		"""
		return tuple(self.prior_params[i] for i in self.all_param_index_to_variable_param_index_map.keys())
		
	@property
	def constant_params(self):
		"""
		The names of the constant parameters
		"""
		return tuple(self.prior_params[i] for i in self.all_param_index_to_constant_param_index_map.keys())
	
	@property
	def consts(self):
		"""
		Returns a dictionary of the constant parameter names and values
		"""
		return dict((p.name,p.const_value) for p in self.constant_params)
	
	def __getitem__(self, k : str | int):
		"""
		Retrieves a parameter by name or index
		"""
		if type(k) == str:
			return self.prior_params[self.param_name_index_map[k]]
		elif type(k) == int:
			return self.prior_params[k]
		else:
			raise IndexError(f'Unknown index type {type(k)} to PriorParamSet')
	
	def append(self, p : PriorParam):
		"""
		Adds a PriorParam to the set 
		"""
		i = len(self.all_param_index_to_variable_param_index_map)
		j = len(self.all_param_index_to_constant_param_index_map)
		k = len(self.prior_params)
		self.prior_params.append(p)
		self.param_name_index_map[p.name] = k
		if not p.is_const:
			self.all_param_index_to_variable_param_index_map[k] = i
		else:
			self.all_param_index_to_constant_param_index_map[k] = j
		return
	
	@staticmethod
	def get_arg_names_of_callable(
			acallable : Callable,
			arg_names : list[str,...] | None = None # names of the arguments to `acallable`, in the correct order as if from inspect.signature(acallable).parameters.keys()
		):
		"""
		Returns the argument names of the passed callable, if the callable has a "arg_names" attribute, returns that otherwise uses the inspect module.
		
		# ARGUMENTS #
			acallable : Callable
				Some callable to get the argument names of
			arg_names : None | list[str,...]
				If not None, returns this instead of doing any work. Useful if you know exactly what the argument names of the callable are.
		"""
		if arg_names is None:
			if hasattr(acallable, 'arg_names'):
				_lgr.debug(f'{acallable.arg_names=}')
				arg_names = acallable.arg_names
			else:
				sig = inspect.signature(acallable)
				_lgr.debug(f'{sig=}')
				arg_names = list(sig.parameters.keys())
		_lgr.debug(f'{arg_names=}')
		return arg_names
	
	def apply_to_callable(self,
			acallable : Callable[P, Any],
			variable_param_values : dict[str,Any],
			const_param_values : dict[str,Any],
			defaults : dict[str, Any] = {},
			not_found_value : Any = None,
			arg_names : list[str,...] | None = None # names of the arguments to `acallable`, in the correct order as if from inspect.signature(acallable).parameters.keys()
		):
		"""
		Given some callable `acallable`, that accepts `arg_names`. Call it with the values for those arguments provided in `variable_param_values`,
		`const_param_values`, `defaults`, `not_found_value`. Where that list is in order of preference (e.g. an argument in both `const_param_values` and `defaults`
		will have the value specified in `const_param_values`.
		
		If `arg_names` is None, will try and infer the argument names of `acallable` via `self.get_arg_names_of_callable()`.
		"""
		_lgr.debug(f'{defaults=}')
		_lgr.debug(f'{const_param_values=}')
		_lgr.debug(f'{variable_param_values=}')
		arg_names = self.get_arg_names_of_callable(acallable, arg_names)
		param_value_dict = {**defaults, **const_param_values, **variable_param_values} # variables overwrite consts which overwrite defaults if there's a conflict
		arg_values = tuple(param_value_dict.get(x,not_found_value) for x in arg_names)

		return acallable(*arg_values)
		
	
	
	def wrap_callable_for_scipy_parameter_order(self, 
			acallable, 
			*,
			arg_to_param_name_map : dict[str,str] = {},
			constant_params_as_defaults=True,
			arg_names : list[str,...] | None = None # names of the arguments to `acallable`, in the correct order as if from inspect.signature(acallable).parameters.keys()
		):
		"""
		Put in a callable with some arguments `callable(arg1, arg2, arg3,...)`, returns a callable 
		that packs all variable params as the first argument, and all const params as the other arguments.
		
		i.e. 
		accepts callable: 
			`callable(arg1, arg2, arg3, arg4, ...)`
		returns callable: 
			`callable((arg1,arg3,...), arg2, arg4, ...)`
		
		# RETURNS #
			new_callable
				A wrapper that accepts all variable parameters as the first argument, and all constant parameters as the other arguments
			var_params
				A list of the variable parameters in the first argument of `new_callable`, in the order they are present in the first argument
			const_params
				A list of the constant parameters that make up the rest of the arguments to `new_callable`, in the order they are present in the argument list.
		"""
		_lgr.debug(f'self.prior_params={tuple(p.name for p in self.prior_params)}')
		_lgr.debug(f'{self.param_name_index_map=}')
		_lgr.debug(f'{self.all_param_index_to_variable_param_index_map=}')
		_lgr.debug(f'{self.all_param_index_to_constant_param_index_map=}')
		
		# acallable example: some_function(carg1, varg2, carg3, varg4)
		arg_names = self.get_arg_names_of_callable(acallable, arg_names)
		"""
		if arg_names is None:
			sig = inspect.signature(acallable)
			_lgr.debug(f'{sig=}')
		
		# prior_params example: [pc3, pc1, pv4, pv2]
		# param_name_index_map: {pc3:0, pc1:1, pv4:2, pv2:3}
		# variable_param_index_map: {2:0, 3:1}
		# constant_param_index_map: {0:0, 1:1}
		
		# arg_to_param_name_map_example: {pc1 : carg1, pv2 : varg2, pc3 : carg3, pv4 : varg4}
		
		#[carg1, varg2, carg3, varg4]
		if arg_names is None:
			arg_names = list(sig.parameters.keys())
		_lgr.debug(f'{arg_names=}')
		"""
		n_args = len(arg_names)
		_lgr.debug(f'{n_args=}')
		
		# {0:1, 1:3, 2:0, 3:2}
		try:
			arg_to_param_ordering = BiDirectionalMap(dict((i,self.param_name_index_map[arg_to_param_name_map.get(arg_name,arg_name)]) for i, arg_name in enumerate(arg_names)))
		except KeyError as e:
			e.add_note(f'NOTE: Could not match up arguments of {acallable.__class__.__name__ if hasattr(acallable,"__call__") else acallable.__name__}({", ".join(arg_names)}), with parameter names {[p.name for p in self.prior_params]} subject to map {arg_to_param_name_map}')
			raise
		
		_lgr.debug(f'{arg_to_param_ordering=}')
		
		# {0:3, 1:4}
		variable_param_order_to_arg_order = dict((self.all_param_index_to_variable_param_index_map[all_param_index], arg_index) for arg_index, all_param_index in arg_to_param_ordering.items() if all_param_index in self.all_param_index_to_variable_param_index_map)
		_lgr.debug(f'{variable_param_order_to_arg_order=}')
		_lgr.debug(f'variable_param_names={tuple(self.param_name_index_map.backward[self.all_param_index_to_variable_param_index_map.backward[var_param_index]] for var_param_index, argument_index in sorted(variable_param_order_to_arg_order.items(), key=lambda x:x[0]))}')
		
		# {0:2, 1:0}
		constant_param_order_to_arg_order = dict((self.all_param_index_to_constant_param_index_map[all_param_index], arg_index) for arg_index, all_param_index in arg_to_param_ordering.items() if all_param_index in self.all_param_index_to_constant_param_index_map)
		_lgr.debug(f'{constant_param_order_to_arg_order=}')
		_lgr.debug(f'constant_param_names={tuple(self.param_name_index_map.backward[self.all_param_index_to_constant_param_index_map.backward[const_param_index]] for const_param_index, argument_index in sorted(constant_param_order_to_arg_order.items(), key=lambda x:x[0]))}')
		
		# Pack information for quick retrieval
		variable_arg_order_array = np.array(tuple(arg_index for var_param_index, arg_index in sorted(variable_param_order_to_arg_order.items(), key=lambda x:x[0])))
		const_arg_order_array = np.array(tuple(arg_index for const_param_index, arg_index in sorted(constant_param_order_to_arg_order.items(), key=lambda x:x[0])))
		
		def new_callable(one_var_arg, *one_const_arg):
			#_lgr.debug(f'{one_var_arg=}')
			#_lgr.debug(f'{one_const_arg=}')
			#_lgr.debug(f'{variable_arg_order_array=}')
			args = [None]*n_args
			for i, j in enumerate(variable_arg_order_array):
				args[j] = one_var_arg[i]
			for i,j in enumerate(const_arg_order_array):
				if i < len(one_const_arg):
					args[j] = one_const_arg[i]
				else:
					args[j] = new_callable.__defaults__[i - len(one_const_arg)]
			return acallable(*args)
		
		if constant_params_as_defaults:
			new_callable.__defaults__ = tuple(self[self.all_param_index_to_constant_param_index_map.backward[const_param_index]].const_value for const_param_index, argument_index in sorted(constant_param_order_to_arg_order.items(), key=lambda x: x[0]))
		_lgr.debug(f'{new_callable.__defaults__=}')
		
		# wrapper function, variable parameter names in order packed into wrapper first arg, constant parameter names in order of rest of wrapper args
		return (
			new_callable, 
			list(self.param_name_index_map.backward[self.all_param_index_to_variable_param_index_map.backward[k]] for k,v in sorted(variable_param_order_to_arg_order.items(), key=lambda x:x[0])), 
			list(self.param_name_index_map.backward[self.all_param_index_to_constant_param_index_map.backward[k]] for k,v in sorted(constant_param_order_to_arg_order.items(), key=lambda x:x[0]))
		)
		
	def wrap_callable_for_ultranest_parameter_order(self, 
			acallable, 
			arg_to_param_name_map : dict[str,str] = {}
		):
		"""
		Put in a callable with some arguments `callable(arg1, arg2, arg3,...)`, returns a callable 
		that packs all params as the first argument.
		
		i.e. 
		accepts callable: 
			`callable(arg1, arg2, arg3, arg4, ...)`
		
		returns callable: 
			`callable(
				(arg1,arg3, arg2, arg4, ...)
			)`
		
		# RETURNS #
			new_callable
				Wrapper that accepts arguments as a single tuple
			param_names
				Parameters names in the order they go into `new_callable`
		"""
		
		# acallable example: some_function(carg1, varg2, carg3, varg4)
		sig = inspect.signature(acallable)
		
		# prior_params example: [pc3, pc1, pv4, pv2]
		# param_name_index_map: {pc3:0, pc1:1, pv4:2, pv2:3}
		# variable_param_index_map: {2:0, 3:1}
		# constant_param_index_map: {0:0, 1:1}
		
		# arg_to_param_name_map_example: {pc1 : carg1, pv2 : varg2, pc3 : carg3, pv4 : varg4}
		
		#[carg1, varg2, carg3, varg4]
		arg_names = list(sig.parameters.keys())
		
		n_args = len(arg_names)
		
		# {1:0, 3:1, 0:2, 2:3}
		param_to_arg_ordering = dict((self.param_name_index_map[arg_to_param_name_map.get(arg_name,arg_name)], i) for i, arg_name in enumerate(arg_names))
		
		def new_callable(all_params):
			args = [None]*n_args
			for i, j in param_to_arg_ordering.items():
				args[j] = all_params[i]
			
			return acallable(*args)
		
		# wrapper function, variable parameter names in order packed into wrapper first arg, constant parameter names in order of rest of wrapper args
		return (
			new_callable, 
			tuple(self.param_name_index_map.backward[k] for k in param_to_arg_ordering.values())
		)
	
	
	
	


"""
Routines useful for geometry
"""
import dataclasses as dc

import numpy as np

def taxicab_metric(p1 : np.ndarray, p2 : np.ndarray, axis : float = -1):
	return np.abs(p1-p2).sum(axis=axis)

def euclidean_metric(p1 : np.ndarray, p2 : np.ndarray, axis : float = -1):
	print(f'{p1.shape} {p2.shape}')
	print(f'{np.sqrt(((p1-p2)**2).sum(axis=axis))}')
	return np.sqrt(((p1-p2)**2).sum(axis=axis))



 
from contextlib import contextmanager

class NotPresent:
	pass

@contextmanager
def attributes(obj, **kwargs):
	"""
	Overwrites the attributes on an object while the context is active. Replaces them with original values on exit.
	"""
	old_attrs = {}
	for k,v in kwargs.items():
		
		if hasattr(obj, k):
			old_attrs[k] = getattr(obj, k)
			setattr(obj, k, v)
		else:
			old_attrs[k] = NotPresent
			if hasattr(obj, "__dict__"):
				setattr(obj, k, v)
			else:
				raise AttributeError(f'Attribute "{k}" not present on object of type "{obj.__class__.__qualname__}", and new attributes cannot be created on this object')
	
	yield obj
	
	for k,v in old_attrs.items():
		if v is NotPresent:
			delattr(obj,k)
		else:
			setattr(obj, k, v)
	



"""
Regrids a numpy array.


Example of grid definition:

	#############################################################

	input_grid    : |---^---|   |---^---|   |---^---|   |---^---|
	              :       |---^---|   |---^---|   |---^---|   	
	
	bin_width     : |-------|

	bin_step      : |-----|

	#############################################################

	output_grid   :     |--------^-------|    |--------^-------|
	              :                |--------^-------|
	
	offset        : |---|

	new_bin_width :     |----------------|
	
	new_bin_step  :     |----------|


	#############################################################

	"|" = bin edge

	"^" = bin center

	width
		The distance between the start and end of a bin
	
	step
		The distance between the start of two consencutive bins

	offset
		The distance between the start/end of the new grid and the start/end of the old grid



"""

import numpy as np

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


def edges_from_midpoints(
		midpoints : np.ndarray,
		fix_edges : bool = True
	) -> np.ndarray:
	"""
	Get bin edges as a 2d-numpy array of shape (2,N), where the
	0th sub-array is the starting point of each bin, and the 1st sub-array is the ending points.

	# ARGUMENTS #
		midpoints : np.ndarray[N]
			Midpoints of bins to find the edges of.
		fix_edges : bool = True
			If we should account for bin size when working out the 0th and Nth bin edges.

	# RETURNS #
		edges : np.ndarray[2,N]
			A 2D numpy array, "edges[0]" holds the starting edges of each bin, 
			"edges[1]" holds the ending edges of each bin.
			
	"""
	if fix_edges:
		x0 = (3*midpoints[0] - midpoints[1])/2
		x1 = (3*midpoints[[-1]] - midpoints[-2])/2
	else:
		x0 = midpoints[0]
		x1 = midpoints[-1]
	edges = np.zeros((2, midpoints.size))
	edges[0,0] = x0
	edges[1,-1] = x1
	_x = (midpoints[:-1] + midpoints[1:])/2
	edges[0,1:] = _x
	edges[1,:-1] = _x

	return edges

def edges_from_bounds(
		lbound : float, 
		ubound : float, 
		steps : float, 
		widths : float, 
		bins_extend_past_ubound=True
	) -> np.ndarray:
	"""
	Get the bin edges of bins defined by a lower bound, upper bound, steps between bins, and widths of each bin.

	# ARGUMENTS #
		lbound : float
			Lower bound of the range to create bins on.
		ubound : float
			Upper bound of the range to create bins on.
		steps : float
			Size of the steps between the starting edge of each bin.
		widths : float
			Size of the bins, i.e. distance between the starting edge and ending edge of each bin.
		bins_extend_past_ubound : bool = True
			Should the bins be allowed to extend past "ubound" if they could contain values < "ubound".
			I.e. Could a bin of width = 5, start at ubound-2 and end at ubound+3?

	# RETURNS #
		edges : np.ndarray[2,N]
			A 2D numpy array, "edges[0]" holds the starting edges of each bin, 
			"edges[1]" holds the ending edges of each bin.
	
	"""
	if not bins_extend_past_ubound:
		ubound = ubound - widths
	_lgr.debug(f'{lbound=} {ubound=} {steps=} {widths=} {bins_extend_past_ubound=}')
	n_bins = int(np.ceil((ubound-lbound)/steps))
	edges = np.full((2,n_bins), fill_value=np.nan)
	edges[0] = np.arange(lbound, ubound, steps)
	edges[1] = edges[0] + widths
	return edges


def regrid(
		data : np.ndarray, 
		input_bin_edges : np.ndarray, 
		output_bin_edges : np.ndarray, 
		axis : int = 0
	) -> np.ndarray:
	"""
	Rebin data from the old bins to a new set of bins, old data is summed into new data bins.
	Set NANs to zero before summing if you want to treat them that way.

	# ARGUMENTS #
		data : np.array[...,N,...]
			The data to be re-gridded. Should make sense to add the data together,
			N is the size of the axis over which data is to be rebinned. Choice of axis
			is controlled by "axis" argument.
		input_bin_edges : np.array[2,N]
			Edges of the old bins, input_bin_edges[0] are the starts of each bin
			input_bin_edges[1] are the ends of each bin.
		output_bin_edges : np.array[2,M]
			Edges of the new bins, output_bin_edges[0] are the starts of each bin,
			output_bin_edges[1] are the ends of each bin.
		axis : int = 0
			Axis to rebin "data" across.

	# RETURNS #
		output_bin_values : np.array[...,M,...]
			Rebinned data values, same shape as "data" except along the rebinned axis.
		n_bins_combined : np.array[M]
			Number of original bins combined into each new bin, fractional bins accounted for.
	"""
	# Move the axis we are working on to the 0th position, that way we
	# don't have to keep track of which axis we are using everywhere. Just remember to
	# move it back at the end.
	np.moveaxis(data, axis, 0)

	i_array= np.arange(input_bin_edges[0].size)
	i_dash_array = np.arange(input_bin_edges[1].size)

	# pos of new bin starts in index of old bin starts
	# floor of this tells us the index of the largest fractional bin at the starting edge
	i_j_array = np.interp(output_bin_edges[0], input_bin_edges[0], i_array)

	# pos of new bin starts in index of old bin ends
	# floor of this tells us the largest index that ends before this new bin starts
	i_dash_j_array = np.interp(output_bin_edges[0], input_bin_edges[1], i_dash_array)
	
	# pos of new bin ends in index of old bin starts
	# floor of this tells us index of the last bin to include in the new bin
	i_j_dash_array = np.interp(output_bin_edges[1], input_bin_edges[0], i_array)
	
	# pos of new bin ends in index of old bin ends
	# floor of this tells us the index of the last full bin included in the new bin
	i_dash_j_dash_array = np.interp(output_bin_edges[1], input_bin_edges[1], i_dash_array)

	# pos of old bin ends in terms of old bin starts
	i_i_dash_array = np.interp(input_bin_edges[1], input_bin_edges[0], i_array)

	output_bin_values = np.zeros((output_bin_edges[0].size, *data.shape[1:]))

	old_bin_idxs = np.arange(input_bin_edges[0].size)
	n_bins_combined = np.zeros((output_bin_edges[0].size,))

	# I feel like there should be a more elegant way to do this.
	for k, (i_j, i_dash_j_dash, i_dash_j, i_j_dash) in enumerate(zip(i_j_array, i_dash_j_dash_array, i_dash_j_array, i_j_dash_array)):
		include_bins = (old_bin_idxs >= i_dash_j) & (old_bin_idxs <= i_j_dash)
		lower_fractional_bins = (old_bin_idxs >= i_dash_j) & (old_bin_idxs < i_j)
		upper_fractional_bins = (old_bin_idxs < i_j_dash) & (old_bin_idxs > i_dash_j_dash)
		whole_bins = (old_bin_idxs >= i_j) & (old_bin_idxs <= i_dash_j_dash)
		whole_bin_idxs = old_bin_idxs[whole_bins]
		lower_fractional_bin_idxs = old_bin_idxs[lower_fractional_bins]
		upper_fractional_bin_idxs = old_bin_idxs[upper_fractional_bins]

		output_bin_values[k] += np.sum(data[whole_bins], axis=0)
		n_bins_combined[k] += np.sum(whole_bins)

		for l, a,b in zip(lower_fractional_bin_idxs, input_bin_edges[0][lower_fractional_bin_idxs], input_bin_edges[1][lower_fractional_bin_idxs]):
			frac_of_bin = (b - output_bin_edges[0][k])/(b-a)
			output_bin_values[k] +=  frac_of_bin* data[l]
			n_bins_combined[k] += frac_of_bin

		for l, a,b in zip(upper_fractional_bin_idxs, input_bin_edges[0][upper_fractional_bin_idxs], input_bin_edges[1][upper_fractional_bin_idxs]):
			frac_of_bin = (output_bin_edges[1][k] - a)/(b-a)
			output_bin_values[k] += frac_of_bin * data[l]
			n_bins_combined[k] += frac_of_bin


	# Once all calculation is complete, move the result back to the
	# axis it is supposed to be in.
	np.moveaxis(output_bin_values, 0, axis)

	return(output_bin_values, n_bins_combined)


#!/usr/bin/env python3
"""
Implementation of Singular Value Decomposition

Mostly consists of helper functions as numpy does heavy lifting. Run the module as __main__
to see and example of it in action.
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

import aopp_deconv_tool.cfg.logs
import aopp_deconv_tool.plot_helper as plot_helper

_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'INFO')


def svd(a : np.ndarray, n : None | int = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
	"""
	Perform single value decomposition on array `a`, see https://en.wikipedia.org/wiki/Singular_value_decomposition
	
	# ARGUMENTS #
		a : np.ndarray
			Array to perform SVD on
		n : int | None
			Number of SVD components to report
	
	# RETURNS #
		evecs : np.ndarray
			"Left side" singular vectors, "U" in A = U S V*
		singular_values : np.ndarray
			Singular values of "left side" and "right side" singular vectors, S in A = U S V*
		fvecs : np.ndarray
			"Right side" singular vectors, V in A = U S V*
	"""
	evals, evecs = np.linalg.eig(a @ a.T)
	e_idxs = np.argsort(evals)[::-1][:n]
	evals = evals[e_idxs]
	evecs = evecs.T[e_idxs].T
	fvecs = a.T @ (evecs/np.sqrt(evals))
	return(evecs, np.diag(np.sqrt(evals)), fvecs.T)


def rect_diag(a, shape):
	"""
	Diagonalise vector a into a matrix of non-rectangular shape
	"""
	d = np.zeros(shape)
	for i in range(min(*shape, a.size)):
		d[i,i]=a[i]
	return(d)

def decompose_to_matricies(u : np.ndarray, s : np.ndarray, v_star : np.ndarray, small : bool = True) -> np.ndarray:
	"""
	Using the results of a singular value decomposition (M = U S V*), returns the matrix M_i =  SUM(s_i * u_i * v_i)
	"""
	if small:
		n = min(s.shape) # number of singular vectors to decompose into
		decomp = np.diag(s)[:,None,None]*(u.T[:n,:n,None] @ v_star[:n,None,:n])
	else:
		decomp = np.diag(s)[:,None,None]*(u.T[:,:,None] @ v_star[:,None,:])
		
	return decomp

def plot(u : np.ndarray, s : np.ndarray, v_star : np.ndarray, inmat : np.ndarray, decomp : np.ndarray, recomp_n=None, f1=None, a1=None):
	"""
	Visualisation of the SVD process, where a matrix M is decomposed into components M =  U S V*
	
	# ARGUMENTS #
		u : np.ndarray
			left singular vectors of some matrix M
		s : np.ndarray
			singular values of some matrix M
		v : np.ndarray
			right singular vectors of some matrix M
		inmat : np.ndarray
			The input matrix, M that `u`, `s`, `v` come from
		decom : np.ndarray
			The M_i decomposition of matrix M, where M =  SUM(M_i) = SUM(s_i * u_i * v_i)
		recomp_n: int | None
			The number of decomposed matrices M_i to recombine for visualisation purposes
		f1 : matplotlib.figure.Figure | None
			Figure to put the plot axes in
		a1 : Sequence[matplotlib.axes.Axes] | None
			6 Axes to plot the data in
	"""
	_lgr.debug(f'{u.shape=}')
	_lgr.debug(f'{s.shape=}')
	_lgr.debug(f'{v_star.shape=}')
	if recomp_n  is None:
		recomp_n = decomp.shape[0]
	recomp = np.sum(decomp[:recomp_n], axis=(0))
	reconstruct = u @ s @ v_star

	_lgr.debug(f'Data ready for plotting')
	
	
	# plot SVD of image
	if f1 is None or a1 is None:
		f1, a1 = plot_helper.figure_n_subplots(6, figure=f1)

	_lgr.debug('figures and axes available')

	a1 = a1.ravel()
	[plot_helper.remove_axes_ticks_and_labels(ax) for ax in a1]

	f1.suptitle('Singular value decomposition U S V* of input maxtrix')

	a1[0].set_title('U matrix')
	a1[0].imshow(u)

	a1[1].set_title('Diagonal elements of S (decending order)')
	a1[1].plot(np.diag(s))
	plot_helper.remove_axes_ticks_and_labels(a1[1], state=True)
	a1[1].set_xlabel('sv idx')
	a1[1].set_ylabel('sv')
	a1[1].axvline(recomp_n, color='red', ls='--')

	a1[2].set_title('V* matrix')
	a1[2].imshow(v_star)

	a1[3].set_title(f'Input matrix\nclim [{np.min(inmat):07.2E},{np.max(inmat):07.2E}]')
	a1[3].imshow(inmat)

	a1[4].set_title(f'Reconstruction from M = U S V*\nclim [{np.min(reconstruct):07.2E},{np.max(reconstruct):07.2E}]')
	a1[4].imshow(reconstruct)

	a1[5].set_title(f'Recomposition from sum(X)_0^{recomp_n}\nclim [{np.min(recomp):07.2E},{np.max(recomp):07.2E}]')
	a1[5].imshow(recomp)


	n_decomp_components_to_plot = 8
	_lgr.debug(f'Plotting {n_decomp_components_to_plot} components of image decomposition')
	# Plot components of image decomposition
	n = min(decomp.shape[0], n_decomp_components_to_plot)
	f2, a2 = plot_helper.figure_n_subplots(n)
	_lgr.debug(f'{f2=} {a2=}')
	a2 = a2.ravel()
	f2.suptitle(f'First {n} svd components X_i [M = sum(X_i)] (of {decomp.shape[0]})')
	for i, ax in enumerate(a2):
		plot_helper.remove_axes_ticks_and_labels(ax)
		ax.set_title(f'i = {i}', y=0.9)
		ax.imshow(decomp[i])
	return


if __name__=='__main__':
	# "arguments"
	recomp_n = None

	# get example data
	try:
		import PIL
		_lgr.info('Creating mandelbrot fractal as test case')
		obs = np.asarray(PIL.Image.effect_mandelbrot((60,50),(0,0,1,1),100))
	except ImportError:
		_lgr.info('Creating random numbers as test case')
		obs = np.random.random((60,50))
	#obs, psf = fitscube.deconvolve.helpers.get_test_data()

	_lgr.info('get singular value decomposition')
	u, s, v_star = np.linalg.svd(obs)

	_lgr.info('get svd from numpy into full form')
	s = rect_diag(s, (u.shape[1],v_star.shape[1]))

	_lgr.info('get decomposed matrices from U S V*')
	decomp = decompose_to_matricies(u, s, v_star)

	_lgr.info('plot singular value decomposition')
	plot(u, s, v_star, obs, decomp, recomp_n)
	plt.show()





#!/usr/bin/env python3
"""
Implements Otsu's method for thresholding images
"""

import aopp_deconv_tool.cfg
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')

import sys, os
import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from astropy.io import fits


def calc(counts, bin_edges):
	"""
	Calculates the inter-class-variance for a histogram with `counts` at all `bin_edges`
	"""
	ns = np.linspace(0, counts.size-1, counts.size)
	inter_class_variance = np.zeros_like(bin_edges, dtype=float)
	for i in range(0,bin_edges.size):
		i1, i2 = ns[:i], ns[i:]
		p1, p2 = counts[:i], counts[i:]
		w1, w2 = np.nansum(p1), np.nansum(p2)
		u1, u2 = np.nansum(p1*i1)/w1, np.nansum(p2*i2)/w2
		inter_class_variance[i] = (w1*w2*(u1-u2)**2)/np.sum(counts)
	return(inter_class_variance)

def threshold(bin_edges, icv):
	"""
	Given the `bin_edges` of a histogram and the inter-class-variance at those edges, returns the edge with the maximum inter-class-variance
	"""
	return(bin_edges[np.nanargmax(icv)] if np.any(~np.isnan(icv)) else np.nan)


def n_thresholds(data, n):
	"""
	Calculates `n` otsu-thresholds for `data`. Automatically calculates the bin edges depending on the sqrt of the number of datapoints
	"""
	ots = np.zeros((n,))
	for i in range(n):
		counts, bin_edges = np.histogram(
			data, 
			bins=np.linspace(np.nanmin(data), np.nanmax(data), int(np.sqrt(np.sum(~np.isnan(data)))))
		)
		icv = calc(counts, bin_edges)
		ots[i] = threshold(bin_edges, icv) if icv.size>2 else np.nan
		data[data<ots[i]] = np.nan
	return(ots)


def frac_per_fpix_threshold(data, frac_per_fpix=15, n_max=10, on_fail='return_last'):
	"""
	Calculates the fraction of pixels included in `n_max` otsu thresholds
	"""
	s = np.nansum(data)
	n = np.sum(~np.isnan(data))
	ot=0
	for i in range(n_max):
		counts, bin_edges = np.histogram(
			data, 
			bins=np.linspace(np.nanmin(data), np.nanmax(data), int(np.sqrt(np.sum(~np.isnan(data)))))
		)
		icv = calc(counts, bin_edges)
		if icv.size <=2:
			return(None if on_fail != 'return_last' else ot)
		ot = threshold(bin_edges, icv)
		data[data<ot] = np.nan
		this_frac_per_fpix = (np.nansum(data)*n)/(np.sum(~np.isnan(data))*s)
		_lgr.debug(i, ot, this_frac_per_fpix)
		if  this_frac_per_fpix > frac_per_fpix:
			return(ot)
	return(None if on_fail != 'return_last' else ot)

def max_frac_per_fpix_threshold(data, n_max=10):
	"""
	Calculates the "most selective" otsu threshold out of `n_max` thresholds
	"""
	s = np.nansum(data)
	n = np.sum(~np.isnan(data))
	ot = np.nanmin(data)
	for i in range(n_max):
		counts, bin_edges = np.histogram(
			data, 
			bins=np.linspace(np.nanmin(data), np.nanmax(data), int(np.sqrt(np.sum(~np.isnan(data)))))
		)
		icv = calc(counts, bin_edges)
		if icv.size <= 2:
			return(ot)
		ot = threshold(bin_edges, icv)
		data[data<ot] = np.nan
		this_frac_per_fpix = (np.nansum(data)*n)/(np.sum(~np.isnan(data))*s)
		_lgr.debug(i, ot, this_frac_per_fpix)
	return(ot)


def max_frac_diff_threshold(data, n_max=10):
	"""
	Returns the otsu threshold (of `n_max` thresholds) that selects a group of similarly valued pixels
	"""
	ots = np.zeros((n_max,))
	frac_diffs = np.zeros((n_max,))
	for i in range(n_max):
		n_valid_px = np.sum(~np.isnan(data))
		counts, bin_edges = np.histogram(
			data, 
			bins=np.linspace(np.nanmin(data), np.nanmax(data), int(np.sqrt(n_valid_px)))
		)
		icv = calc(counts, bin_edges)
		ots[i] = threshold(bin_edges, icv) if icv.size>2 else np.nan
		data[data<ots[i]] = np.nan

		range_frac = (ots[i]-bin_edges[0])/(bin_edges[-1]-bin_edges[0])
		npix_frac = np.sum(~np.isnan(data))/n_valid_px
		frac_diffs[i] = (range_frac-npix_frac)/(range_frac+npix_frac)
		if np.all(np.isnan(frac_diffs)):
			return(np.nan)

	
	return(ots[np.nanargmax(frac_diffs)])



def polynomial(coef, x):
	return(np.nansum(coef[:,None]*np.vstack([x**i for i in range(len(coef)-1,-1,-1)]),axis=0))


def plot_repeats(data, n=4):
	countsl, bin_edgesl, bin_midsl, icvl, otl, rdatal, odatal = [], [], [], [], [], [], []
	polyfits = []

	odatal.append(data)
	for i in range(n):
		counts, bin_edges = np.histogram(
			odatal[i], 
			bins=np.linspace(np.nanmin(odatal[i]), np.nanmax(odatal[i]), int(np.sqrt(np.sum(~np.isnan(odatal[i]))))),
			density=False
		)
		icv = calc(counts, bin_edges)
		if icv.size<=2:
			ot = np.nan
		else:
			ot = threshold(bin_edges, icv)
	
		rdata = np.array(odatal[i])
		rdata[rdata<ot] = np.nan
	
		range_frac = (ot-bin_edges[0])/(bin_edges[-1]-bin_edges[0])
		npix_frac = np.sum(~np.isnan(rdata))/np.sum(~np.isnan(odatal[i]))
		_lgr.debug(range_frac, npix_frac, (range_frac-npix_frac)/(range_frac+npix_frac))

		countsl.append(counts)
		bin_edgesl.append(bin_edges)
		bin_midsl.append(0.5*(bin_edges[1:]+bin_edges[:1]))
		icvl.append(icv)
		otl.append(ot)
		rdatal.append(rdata)
		odatal.append(rdata)

	

	ots = n_thresholds(np.array(data), n)

	f1, axes = ut.plt.create_figure_with_subplots(4, n, sp_kwargs={'gridspec_kw':{'hspace':0.3}})
	axes_iter = iter(axes.T.flatten())

	s = np.nansum(data)
	pn = np.sum(~np.isnan(data))

	for i in range(n):	

		ax = next(axes_iter)
		ax.imshow(odatal[i])
		ax.set_title(f'frac {np.nansum(odatal[i])/s:08.2E}\nfrac/fpx {(np.nansum(odatal[i])*pn)/(s*np.sum(~np.isnan(odatal[i]))):08.2E}\nfrac^2/fpx {(pn*np.nansum(odatal[i])**2)/(np.sum(~np.isnan(odatal[i]))*s**2):08.2E}')
		ut.plt.remove_axes_ticks_and_labels(ax)
		
		ax = next(axes_iter)
		ax.step(bin_edgesl[i][1:], countsl[i])
		ax.axvline(otl[i], color='red')
		ax.axvline(ots[i], color='green', ls='--')


		ax = next(axes_iter)
		ax.plot(bin_edgesl[i], icvl[i])
		ax.axvline(otl[i], color='red')
		ax.axvline(ots[i], color='green', ls='--')

		ax = next(axes_iter)
		ax.imshow(rdatal[i])
		ax.set_title(f'frac {np.nansum(rdatal[i])/s:08.2E}\nfrac/fpx {(np.nansum(odatal[i])*pn)/(s*np.sum(~np.isnan(odatal[i]))):08.2E}')
		ut.plt.remove_axes_ticks_and_labels(ax)

	

	f1.show()
	input("Close figure and press [ENTER] to continue > ")
	plt.close(f1)
	
	


if __name__=='__main__':

	fpath = os.path.expanduser("~/scratch/general_testing/telecope_data/obs10_MUSE.2019-10-18T00:32:18.867/DATACUBE_FINAL_SMOOTH_10.fits")
	fpath = os.path.expanduser("~/scratch/general_testing/test_archive/Neptune/VLT_MUSE/2018/20180619_1_V/Neptune_MUSE_20180619_1_V_ALTERED.fits")


	with fits.open(fpath) as hdul:
		hdul.info()
		#data_idx = 440
		data_idx = 26
		data = np.abs(hdul[1].data[data_idx])

		plot_repeats(data, 10)

		_lgr.DEBUG(f'{frac_per_fpix_threshold(np.array(data), 15)=}')
		_lgr.DEBUG(f'{max_frac_per_fpix_threshold(np.array(data), 20)=}')









"""
Provide various ways of estimating the noise in some data
"""


import itertools as it
import numpy as np

import aopp_deconv_tool.cfg.logs
_lgr = aopp_deconv_tool.cfg.logs.get_logger_at_level(__name__, 'DEBUG')


def corners_standard_deviation(a : np.ndarray, corner_frac : float = 1/10) -> float:
	"""
	Uses the "corners" (i.e. extremes in all directions) of array `a` to estimate the noise. This 
	assumes our signal does not extend all the way to the edges. 
	
	# ARGUMENTS #
		a : np.ndarray
			Input array to estimate noise of
		corner_frac : float = 1/10
			Fraction of the range of `a` that is considered a "corner"
	
	# RETURNS #
		float - Standard deviation of the "corners" of `a`
	"""
	corner_mask = np.zeros_like(a, dtype=bool)
	low_corner_slices = tuple(slice(0, int(np.floor(s*corner_frac)), None) for s in a.shape) # xl, yl, zl
	high_corner_slices = tuple(slice(-int(np.floor(s*corner_frac)), None, None) for s in a.shape) # xh, yh, zh
	
	for corner in it.product(*zip(low_corner_slices, high_corner_slices)):
		corner_mask[corner] = 1
	
	return np.std(a[corner_mask])